#include "Game.hpp"
#include <algorithm>

using std::string;
using std::vector;

Game::Game() :
    m_punter(0),
    m_numPunters(0),
    m_useFutures(false),
    m_useSplurges(false),
    m_useOptions(false)
{
}

Game::~Game()
{
}

void Game::initOptionsLeft()
{
    if (!m_useOptions)
    {
        return;
    }

    m_optionsLeft.resize(m_numPunters);
    for (uint32_t iPunter = 0; iPunter < m_numPunters; iPunter++)
    {
        m_optionsLeft[iPunter] = m_map.m_numMines;
    }
    for (uint32_t iRiver = 0; iRiver < m_map.m_numRivers; iRiver++)
    {
        uint32_t punter = m_options[iRiver];
        if (punter != (uint32_t)-1)
        {
            if (m_optionsLeft[punter] > 0)
            {
                m_optionsLeft[punter]--;
            }
        }
    }
}

bool Game::checkAndApply(const vector<uint32_t>& route,
                         bool bApply,
                         string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    uint32_t routeSize = (uint32_t)route.size();
    if (routeSize == 1)
    {
        *pMsg = "Splurges not enabled";
        return false;
    }

    if (routeSize > 2 && !m_useSplurges)
    {
        *pMsg = "One site";
        return false;
    }

    for (uint32_t iSite = 0; iSite < routeSize; iSite++)
    {
        if (route[iSite] >= m_map.m_numSites)
        {
            *pMsg = "Site index out of range";
            return false;
        }
    }

    uint32_t optionsLeft = 0;
    if (m_useOptions)
    {
        optionsLeft = m_optionsLeft[m_punter];
    }

    vector<uint32_t> routeRivers;
    for (uint32_t iSite = 1; iSite < routeSize; iSite++)
    {
        River river(route[iSite - 1], route[iSite]);
        uint32_t riverIdx = 0;
        if (!m_map.getRiverIdx(river, &riverIdx))
        {
            *pMsg = "River index not found";
            return false;
        }
        if (m_claims[riverIdx] != (uint32_t)-1)
        {
            if (!m_useOptions)
            {
                *pMsg = "River already claimed";
                return false;
            }
            if (optionsLeft == 0)
            {
                *pMsg = "No options left";
                return false;
            }
            optionsLeft--;
        }
        routeRivers.push_back(riverIdx);
    }

    uint32_t riversSize = (uint32_t)routeRivers.size();
    if (!routeRivers.empty())
    {
        vector<uint32_t> sortedRivers(routeRivers);
        std::sort(sortedRivers.begin(), sortedRivers.end());
        for (uint32_t iRiver = 1; iRiver < riversSize; iRiver++)
        {
            if (sortedRivers[iRiver - 1] == sortedRivers[iRiver])
            {
                *pMsg = "Duplicate river";
                return false;
            }
        }
    }

    if (bApply)
    {
        for (uint32_t riverIdx : routeRivers)
        {
            if (m_claims[riverIdx] == (uint32_t)-1)
            {
                m_claims[riverIdx] = m_punter;
            }
            else
            {
                m_options[riverIdx] = m_punter;
                m_optionsLeft[m_punter]--;
            }
        }
    }

    return true;
}

void Game::serialize(vector<uint32_t>* pData) const
{
    auto& data = *pData;

    m_map.serialize(pData);

    data.push_back(m_punter);
    data.push_back(m_numPunters);

    uint32_t settings = 0;
    if (m_useFutures) settings |= 0x1;
    if (m_useSplurges) settings |= 0x2;
    if (m_useOptions) settings |= 0x4;
    data.push_back(settings);

    for (uint32_t iRiver = 0; iRiver < m_map.m_numRivers; iRiver++)
    {
        data.push_back(m_claims[iRiver]);
    }

    if (m_useOptions)
    {
        for (uint32_t iRiver = 0; iRiver < m_map.m_numRivers; iRiver++)
        {
            data.push_back(m_options[iRiver]);
        }
    }
}

bool Game::deserialize(const vector<uint32_t>& data, size_t& pos)
{
    size_t size = data.size();

    if (!m_map.deserialize(data, pos))
    {
        return false;
    }

    if (pos + 3 + m_map.m_numRivers > size)
    {
        return false;
    }

    m_punter = data[pos++];
    m_numPunters = data[pos++];

    uint32_t settings = data[pos++];
    m_useFutures = ((settings & 0x1) != 0);
    m_useSplurges = ((settings & 0x2) != 0);
    m_useOptions = ((settings & 0x4) != 0);

    m_claims.resize(m_map.m_numRivers);
    for (uint32_t iRiver = 0; iRiver < m_map.m_numRivers; iRiver++)
    {
        m_claims[iRiver] = data[pos++];
    }

    if (m_useOptions)
    {
        if (pos + m_map.m_numRivers > size)
        {
            return false;
        }

        m_options.resize(m_map.m_numRivers);
        for (uint32_t iRiver = 0; iRiver < m_map.m_numRivers; iRiver++)
        {
            m_options[iRiver] = data[pos++];
        }
    }

    initOptionsLeft();

    return true;
}
