#include "GreedySolver.hpp"
#include "Game.hpp"
#include "Map.hpp"
#include <random>

using std::vector;

#define VERBOSE 0

GreedySolver::GreedySolver() :
    m_haveFirstMove(false),
    m_firstSource((uint32_t)-1),
    m_firstTarget((uint32_t)-1)
{
}

GreedySolver::~GreedySolver()
{
}

void GreedySolver::solve(const Game& game,
                         vector<uint32_t>* pRoute)
{
    if (m_haveFirstMove)
    {
        if (m_firstSource != (uint32_t)-1 &&
            m_firstTarget != (uint32_t)-1)
        {
            pRoute->push_back(m_firstSource);
            pRoute->push_back(m_firstTarget);
        }
        m_haveFirstMove = false;
        m_firstSource = (uint32_t)-1;
        m_firstTarget = (uint32_t)-1;
        return;
    }

    auto& map = game.m_map;
    uint32_t numSites = map.m_numSites;
    uint32_t numRivers = map.m_numRivers;
    uint32_t numMines = map.m_numMines;
    uint32_t punter = game.m_punter;

    uint32_t optionsLeft = 0;
    if (game.m_useOptions)
    {
        optionsLeft = game.m_optionsLeft[game.m_punter];
    }

    uint32_t numComponents = 0;
    vector<uint32_t> siteComponents(numSites);
    for (uint32_t iSite = 0; iSite < numSites; iSite++)
    {
        siteComponents[iSite] = (uint32_t)-1;
    }

    vector<uint32_t> stack;
    for (uint32_t iSite = 0; iSite < numSites; iSite++)
    {
        if (siteComponents[iSite] != (uint32_t)-1)
        {
            continue;
        }

        stack.push_back(iSite);
        while (!stack.empty())
        {
            uint32_t iCurSite = stack.back();
            stack.pop_back();

            if (siteComponents[iCurSite] != (uint32_t)-1)
            {
                continue;
            }
            siteComponents[iCurSite] = numComponents;

            uint32_t low = map.m_siteRiverBounds[iCurSite].m_low;
            uint32_t high = map.m_siteRiverBounds[iCurSite].m_high;
            for (uint32_t iSiteRiver = low; iSiteRiver < high; iSiteRiver++)
            {
                auto& siteRiver = map.m_siteRivers[iSiteRiver];
                if (!(game.m_claims[siteRiver.m_river] == punter ||
                      (game.m_useOptions &&
                       game.m_options[siteRiver.m_river] == punter)))
                {
                    continue;
                }

                uint32_t target = siteRiver.m_target;
                if (siteComponents[target] == (uint32_t)-1)
                {
                    stack.push_back(target);
                }
            }
        }

        numComponents++;
    }

#if VERBOSE
    fprintf(stderr, "numComponents = %" PRIu32 "\n", numComponents);
#endif

    uint32_t numMineComponents = 0;
    vector<uint32_t> componentMineComponents(numComponents);
    for (uint32_t iComponent = 0; iComponent < numComponents; iComponent++)
    {
        componentMineComponents[iComponent] = (uint32_t)-1;
    }
    for (uint32_t iMine = 0; iMine < numMines; iMine++)
    {
        uint32_t iSite = map.m_mines[iMine];
        uint32_t iComponent = siteComponents[iSite];
        if (componentMineComponents[iComponent] != (uint32_t)-1)
        {
            continue;
        }

        componentMineComponents[iComponent] = numMineComponents;
        numMineComponents++;
    }

#if VERBOSE
    fprintf(stderr, "numMineComponents = %" PRIu32 "\n", numMineComponents);
#endif

    vector<uint32_t> scoreTable(numMineComponents * numComponents);

    vector<uint32_t> queueCurrent;
    vector<uint32_t> queueNext;
    vector<uint8_t> visited(numSites);
    for (uint32_t iMine = 0; iMine < numMines; iMine++)
    {
        uint32_t iSite = map.m_mines[iMine];
        uint32_t iComponent = siteComponents[iSite];
        uint32_t iMineComponent = componentMineComponents[iComponent];

        for (uint32_t iSite = 0; iSite < numSites; iSite++)
        {
            visited[iSite] = 0;
        }

        uint32_t dist = 0;
        queueNext.push_back(iSite);
        while (!queueNext.empty())
        {
            queueCurrent.swap(queueNext);
            for (uint32_t iCurSite : queueCurrent)
            {
                if (visited[iCurSite])
                {
                    continue;
                }
                visited[iCurSite] = 1;
                uint32_t iCurComponent = siteComponents[iCurSite];
                if (iCurComponent != iComponent)
                {
                    scoreTable[iMineComponent * numComponents + iCurComponent]
                        += dist * dist;
                }

                uint32_t low = map.m_siteRiverBounds[iCurSite].m_low;
                uint32_t high = map.m_siteRiverBounds[iCurSite].m_high;
                for (uint32_t iSiteRiver = low; iSiteRiver < high; iSiteRiver++)
                {
                    auto& siteRiver = map.m_siteRivers[iSiteRiver];
                    uint32_t target = siteRiver.m_target;
                    if (!visited[target])
                    {
                        queueNext.push_back(target);
                    }
                }
            }
            queueCurrent.clear();
            dist++;
        }
    }

    vector<uint32_t> siteLiberties(numSites);
    for (uint32_t iRiver = 0; iRiver < numRivers; iRiver++)
    {
        if (game.m_claims[iRiver] != (uint32_t)-1)
        {
            continue;
        }

        auto& river = map.m_rivers[iRiver];
        siteLiberties[river.m_source]++;
        siteLiberties[river.m_target]++;
    }

    uint32_t bestMove = (uint32_t)-1;
    uint32_t bestScore = 0;
    uint32_t bestOptionsLeft = 0;
    uint32_t bestLiberties = 0;
    uint32_t bestCount = 0;

    std::default_random_engine gen{std::random_device()()};

    for (uint32_t iRiver = 0; iRiver < numRivers; iRiver++)
    {
        if (!(game.m_claims[iRiver] == (uint32_t)-1 ||
              (optionsLeft > 0 &&
               game.m_claims[iRiver] != game.m_punter &&
               game.m_options[iRiver] == (uint32_t)-1)))
        {
            continue;
        }

        auto& river = map.m_rivers[iRiver];
        uint32_t iSourceComponent = siteComponents[river.m_source];
        uint32_t iTargetComponent = siteComponents[river.m_target];
        uint32_t iSourceMineComponent = componentMineComponents[iSourceComponent];
        uint32_t iTargetMineComponent = componentMineComponents[iTargetComponent];

        uint32_t curScore = 0;
        if (iSourceMineComponent != (uint32_t)-1)
        {
            curScore += scoreTable[iSourceMineComponent * numComponents + iTargetComponent];
        }
        if (iTargetMineComponent != (uint32_t)-1)
        {
            curScore += scoreTable[iTargetMineComponent * numComponents + iSourceComponent];
        }

        uint32_t curOptionsLeft = optionsLeft;
        if (game.m_claims[iRiver] != (uint32_t)-1)
        {
            curOptionsLeft--;
        }

        uint32_t curLiberties = 0;
        curLiberties += siteLiberties[river.m_source];
        curLiberties += siteLiberties[river.m_target];

#if VERBOSE
        fprintf(stderr, "%4" PRIu32 ": %" PRIu32 " (%" PRIu32 " liberties)\n",
                iRiver,
                curScore,
                curLiberties);
#endif

        if (bestMove == (uint32_t)-1 ||
            curScore > bestScore ||
            (curScore == bestScore &&
             curOptionsLeft > bestOptionsLeft) ||
            (curScore == bestScore &&
             curOptionsLeft == bestOptionsLeft &&
             curLiberties > bestLiberties))
        {
            bestMove = iRiver;
            bestScore = curScore;
            bestOptionsLeft = curOptionsLeft;
            bestLiberties = curLiberties;
            bestCount = 1;
        }
        else if (curScore == bestScore &&
                 curOptionsLeft == bestOptionsLeft &&
                 curLiberties == bestLiberties)
        {
            bestCount++;
            std::uniform_int_distribution<uint32_t> dist(0, bestCount - 1);
            if (dist(gen) == 0)
            {
                bestMove = iRiver;
            }
        }
    }

    if (bestMove == (uint32_t)-1)
    {
        // No legal moves -- should not happen
        return; // pass
    }

    {
#if VERBOSE
        fprintf(stderr, "Best:\n");
        fprintf(stderr, "%4" PRIu32 ": %" PRIu32 " (%" PRIu32 " liberties)\n",
                bestMove,
                bestScore,
                bestLiberties);
#endif
        auto& river = map.m_rivers[bestMove];
        pRoute->push_back(river.m_source);
        pRoute->push_back(river.m_target);
        return;
    }
}

void GreedySolver::getFutures(const Game& game,
                              vector<SolverFuture>* pFutures)
{
    if (game.m_punter != 0)
    {
        return;
    }

    vector<uint32_t> route;
    solve(game, &route);

    m_haveFirstMove = true;

    if (route.size() != 2)
    {
        m_firstSource = (uint32_t)-1;
        m_firstTarget = (uint32_t)-1;
        return;
    }

    m_firstSource = route[0];
    m_firstTarget = route[1];

    if (game.m_map.m_isMine[m_firstSource] &&
        !game.m_map.m_isMine[m_firstTarget])
    {
        pFutures->emplace_back(m_firstSource, m_firstTarget);
    }
    else if (game.m_map.m_isMine[m_firstTarget] &&
             !game.m_map.m_isMine[m_firstSource])
    {
        pFutures->emplace_back(m_firstTarget, m_firstSource);
    }
}

void GreedySolver::serialize(vector<uint32_t>* pData) const
{
    pData->push_back(m_haveFirstMove ? 1 : 0);
    pData->push_back(m_firstSource);
    pData->push_back(m_firstTarget);
}

bool GreedySolver::deserialize(const vector<uint32_t>& data,
                               size_t& pos)
{
    size_t size = data.size();
    if (pos + 3 > size)
    {
        return false;
    }

    m_haveFirstMove = (data[pos++] != 0);
    m_firstSource = data[pos++];
    m_firstTarget = data[pos++];
    return true;
}
