#include "Map.hpp"
#include <algorithm>

using std::vector;

Map::Map() :
    m_numSites(0),
    m_numRivers(0),
    m_numMines(0)
{
}

Map::~Map()
{
}

uint32_t Map::getSiteId(uint32_t siteIdx) const
{
    return m_siteIds[siteIdx];
}

bool Map::getSiteIdx(uint32_t siteId, uint32_t* pSiteIdx) const
{
    auto it = std::lower_bound(m_siteIds.begin(), m_siteIds.end(), siteId);
    if (it == m_siteIds.end() ||
        *it != siteId)
    {
        return false;
    }

    if (pSiteIdx)
    {
        *pSiteIdx = (uint32_t)(it - m_siteIds.begin());
    }
    return true;
}

bool Map::getRiverIdx(const River& river, uint32_t* pRiverIdx) const
{
    auto it = std::lower_bound(m_rivers.begin(), m_rivers.end(), river);
    if (it == m_rivers.end() ||
        *it != river)
    {
        return false;
    }

    if (pRiverIdx)
    {
        *pRiverIdx = (uint32_t)(it - m_rivers.begin());
    }
    return true;
}

void Map::initSiteRivers()
{
    m_siteRivers.resize(m_numRivers * 2);
    for (uint32_t iRiver = 0; iRiver < m_numRivers; iRiver++)
    {
        m_siteRivers[iRiver * 2].m_source = m_rivers[iRiver].m_source;
        m_siteRivers[iRiver * 2].m_target = m_rivers[iRiver].m_target;
        m_siteRivers[iRiver * 2].m_river = iRiver;
        m_siteRivers[iRiver * 2 + 1].m_source = m_rivers[iRiver].m_target;
        m_siteRivers[iRiver * 2 + 1].m_target = m_rivers[iRiver].m_source;
        m_siteRivers[iRiver * 2 + 1].m_river = iRiver;
    }
    std::sort(m_siteRivers.begin(), m_siteRivers.end());

    m_siteRiverBounds.resize(m_numSites);
    uint32_t source = 0;
    m_siteRiverBounds[0].m_low = 0;
    for (uint32_t iSiteRiver = 0; iSiteRiver < m_numRivers * 2; iSiteRiver++)
    {
        uint32_t curSource = m_siteRivers[iSiteRiver].m_source;
        while (source < curSource)
        {
            m_siteRiverBounds[source].m_high = iSiteRiver;
            source++;
            m_siteRiverBounds[source].m_low = iSiteRiver;
        }
    }
    while (true)
    {
        m_siteRiverBounds[source].m_high = m_numRivers * 2;
        source++;
        if (source >= m_numSites)
        {
            break;
        }
        m_siteRiverBounds[source].m_low = m_numRivers * 2;
    }
}

void Map::initIsMine()
{
#if 1
    m_isMine.resize(m_numSites);
    for (uint32_t iSite = 0; iSite < m_numSites; iSite++)
    {
        m_isMine[iSite] = 0;
    }
    for (uint32_t iMine = 0; iMine < m_numMines; iMine++)
    {
        m_isMine[m_mines[iMine]] = 1;
    }
#endif
}

void Map::serialize(vector<uint32_t>* pData) const
{
    auto& data = *pData;
    data.push_back(m_numSites);
    data.push_back(m_numRivers);
    data.push_back(m_numMines);

    for (uint32_t iSite = 0; iSite < m_numSites; iSite++)
    {
        data.push_back(m_siteIds[iSite]);
    }

    for (uint32_t iRiver = 0; iRiver < m_numRivers; iRiver++)
    {
        data.push_back(m_rivers[iRiver].m_source);
        data.push_back(m_rivers[iRiver].m_target);
    }

    for (uint32_t iMine = 0; iMine < m_numMines; iMine++)
    {
        data.push_back(m_mines[iMine]);
    }
}

bool Map::deserialize(const vector<uint32_t>& data, size_t& pos)
{
    size_t size = data.size();
    if (pos + 3 > size)
    {
        return false;
    }

    m_numSites = data[pos++];
    m_numRivers = data[pos++];
    m_numMines = data[pos++];

    if (pos + m_numSites + m_numRivers * 2 + m_numMines > size)
    {
        return false;
    }

    m_siteIds.resize(m_numSites);
    for (uint32_t iSite = 0; iSite < m_numSites; iSite++)
    {
        m_siteIds[iSite] = data[pos++];
    }

    m_rivers.resize(m_numRivers);
    for (uint32_t iRiver = 0; iRiver < m_numRivers; iRiver++)
    {
        m_rivers[iRiver].m_source = data[pos++];
        m_rivers[iRiver].m_target = data[pos++];
    }

    initSiteRivers();

    m_mines.resize(m_numMines);
    for (uint32_t iMine = 0; iMine < m_numMines; iMine++)
    {
        m_mines[iMine] = data[pos++];
    }

    initIsMine();

    m_minCoord.m_x = 0.0;
    m_minCoord.m_y = 0.0;
    m_maxCoord.m_x = 0.0;
    m_maxCoord.m_y = 0.0;

    m_siteCoords.resize(m_numSites);
    for (uint32_t iSite = 0; iSite < m_numSites; iSite++)
    {
        auto& siteCoord = m_siteCoords[iSite];
        siteCoord.m_x = 0.0;
        siteCoord.m_y = 0.0;
    }

    return true;
}
