#include "NetUtils.hpp"
#include "Cleanup.hpp"
#include "TimeUtils.hpp"
#include "StringUtils.hpp"

#if PLATFORM_WINDOWS
#define IPV6STRICT
#include <winsock2.h>
#include <windows.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/select.h>
#endif

using std::vector;
using std::string;

#if PLATFORM_WINDOWS
typedef SOCKET OSSocket;
#define OSInvalidSocket INVALID_SOCKET
#define OSSocketError SOCKET_ERROR
#else
typedef int OSSocket;
#define OSInvalidSocket (-1)
#define OSSocketError (-1)
#endif

namespace
{
    int getNetworkError()
    {
#if PLATFORM_WINDOWS
        return WSAGetLastError();
#else
        return errno;
#endif
    }

    string getNetworkErrorMsg(int err)
    {
#if PLATFORM_WINDOWS
        size_t bufSize = 32000;
        char buf[bufSize];
        if (FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM |
                           FORMAT_MESSAGE_IGNORE_INSERTS,
                           nullptr,
                           (DWORD)err,
                           MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                           buf,
                           (DWORD)(sizeof(buf) / sizeof(buf[0])),
                           nullptr) != 0)
        {
            size_t iOut = 0;
            for (size_t iIn = 0; iIn < bufSize; iIn++)
            {
                if (buf[iIn] >= 32 && buf[iIn] <= 126)
                {
                    buf[iOut++] = buf[iIn];
                }
                else if (buf[iIn] == '\0')
                {
                    buf[iOut++] = buf[iIn];
                    break;
                }
            }
            return buf;
        }
#else
        char* strerrMsg = strerror(err);
        if (strerrMsg)
        {
            return strerrMsg;
        }
#endif
        return strprintf("Error %d", err);
    }

#if PLATFORM_WINDOWS
    void shutdownWrite(OSSocket osSocket) { shutdown(osSocket, SD_SEND); }
    void shutdownRead(OSSocket osSocket) { shutdown(osSocket, SD_RECEIVE); }
    void shutdownBoth(OSSocket osSocket) { shutdown(osSocket, SD_BOTH); }
    void closeSocket(OSSocket osSocket) { closesocket(osSocket); }
#else
    void shutdownWrite(OSSocket osSocket) { shutdown(osSocket, SHUT_WR); }
    void shutdownRead(OSSocket osSocket) { shutdown(osSocket, SHUT_RD); }
    void shutdownBoth(OSSocket osSocket) { shutdown(osSocket, SHUT_RDWR); }
    void closeSocket(OSSocket osSocket) { close(osSocket); }
#endif

    bool setNonBlocking(OSSocket osSocket, string* pMsg = nullptr)
    {
#if PLATFORM_WINDOWS
        u_long nOne = 1;
        int result = ioctlsocket(osSocket, FIONBIO, &nOne);
#else
        int flags = fcntl(osSocket, F_GETFL, 0);
        if (flags == OSSocketError)
        {
            flags = 0;
        }
        int result = fcntl(osSocket, F_SETFL, flags | O_NONBLOCK);
#endif
        if (result == OSSocketError)
        {
            if (pMsg)
            {
                int err = getNetworkError();
#if PLATFORM_WINDOWS
                *pMsg = "ioctlsocket error: ";
#else
                *pMsg = "fcntl error: ";
#endif
                *pMsg += getNetworkErrorMsg(err);
            }
            return false;
        }

        return true;
    }
}

bool networkStartup(string* pMsg)
{
#if PLATFORM_WINDOWS
    WORD wVersionRequested = MAKEWORD(2, 2);
    WSADATA wsaData;
    int result = WSAStartup(wVersionRequested, &wsaData);
    if (result != 0)
    {
        if (pMsg)
        {
            *pMsg = getNetworkErrorMsg(result);
        }
        return false;
    }
    if (wsaData.wVersion != wVersionRequested)
    {
        WSACleanup();
        if (pMsg)
        {
            *pMsg = "Winsock version not supported";
        }
        return false;
    }
#endif
    return true;
}

bool networkCleanup(string* pMsg)
{
#if PLATFORM_WINDOWS
    int result = WSACleanup();
    if (result == OSSocketError)
    {
        if (pMsg)
        {
            int err = getNetworkError();
            *pMsg = getNetworkErrorMsg(err);
        }
        return false;
    }
#endif
    return true;
}

bool lookupAddresses(const string& hostName,
                     const string& portName,
                     vector<Address>* pAddresses,
                     string* pMsg)
{
    struct addrinfo hints;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags = AI_ADDRCONFIG;

    struct addrinfo* res = nullptr;
    int err =
        getaddrinfo(hostName.c_str(),
                    portName.c_str(),
                    &hints,
                    &res);

    if (err != 0)
    {
        if (pMsg)
        {
            *pMsg = "getaddrinfo error: ";
#if PLATFORM_WINDOWS
            *pMsg += getNetworkErrorMsg(getNetworkError());
#else
            *pMsg += gai_strerror(err);
#endif
        }
        return false;
    }

    Cleanup resCleanup([&]{ freeaddrinfo(res); });

    vector<Address> addresses;
    for (struct addrinfo* p = res; p != nullptr; p = p->ai_next)
    {
        if (p->ai_family ==
#if PLATFORM_WINDOWS
            2
#else
            AF_INET
#endif
            ||
            p->ai_family == AF_INET6)
        {
            char hostBuf[NI_MAXHOST];
            char portBuf[NI_MAXSERV];
            if (getnameinfo(p->ai_addr, p->ai_addrlen,
                            hostBuf, sizeof(hostBuf),
                            portBuf, sizeof(portBuf),
                            NI_NUMERICHOST | NI_NUMERICSERV) == 0)
            {
                addresses.push_back(Address(hostBuf, portBuf));
            }
        }
    }

    if (addresses.empty())
    {
        if (pMsg)
        {
            *pMsg = "No addresses found";
        }
        return false;
    }

    if (pAddresses)
    {
        pAddresses->swap(addresses);
    }

    return true;
}

class TCPSocketImpl
{
public:
    OSSocket m_osSocket;
    SocketState m_state;
    string m_msg;
    uint64_t m_lastSendTimeMS;
    uint64_t m_lastRecvTimeMS;
    bool m_shutdownRequested;
};

TCPSocket::TCPSocket() :
    m_pImpl(new TCPSocketImpl)
{
    m_pImpl->m_osSocket = OSInvalidSocket;
    m_pImpl->m_state = SocketState::Disconnected;
    m_pImpl->m_lastSendTimeMS = 0;
    m_pImpl->m_lastRecvTimeMS = 0;
    m_pImpl->m_shutdownRequested = false;
}

TCPSocket::TCPSocket(TCPSocket&& other) :
    m_pImpl(std::move(other.m_pImpl))
{
    other.m_pImpl.reset(new TCPSocketImpl);
}

TCPSocket::~TCPSocket()
{
    reset();
}

TCPSocket& TCPSocket::operator=(TCPSocket&& other)
{
    if (this != &other)
    {
        reset();
        m_pImpl = std::move(other.m_pImpl);
        other.m_pImpl.reset(new TCPSocketImpl);
    }
    return *this;
}

SocketState TCPSocket::getState() const
{
    return m_pImpl->m_state;
}

string TCPSocket::getMsg() const
{
    return m_pImpl->m_msg;
}

void TCPSocket::connect(const Address& address)
{
    if (m_pImpl->m_state != SocketState::Disconnected)
    {
        reset();
    }

    struct addrinfo hints;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags =
        AI_NUMERICHOST |
#if PLATFORM_WINDOWS
        8
#else
        AI_NUMERICSERV
#endif
        ;

    struct addrinfo* res = nullptr;

    {
        int err =
            getaddrinfo(address.m_host.c_str(),
                        address.m_port.c_str(),
                        &hints,
                        &res);

        if (err != 0)
        {
            m_pImpl->m_state = SocketState::Error;
            m_pImpl->m_msg = "getaddrinfo error: ";
#if PLATFORM_WINDOWS
            m_pImpl->m_msg += getNetworkErrorMsg(getNetworkError());
#else
            m_pImpl->m_msg += gai_strerror(err);
#endif
            return;
        }
    }

    if (!res)
    {
        m_pImpl->m_state = SocketState::Error;
        m_pImpl->m_msg = "getaddrinfo returned no results";
        return;
    }

    Cleanup resCleanup([&]{ freeaddrinfo(res); });

    OSSocket osSocket =
        socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (osSocket == OSInvalidSocket)
    {
        m_pImpl->m_state = SocketState::Error;
        int err = getNetworkError();
        m_pImpl->m_msg = "socket error: ";
        m_pImpl->m_msg += getNetworkErrorMsg(err);
        return;
    }

    Cleanup socketCleanup([&]{ closeSocket(osSocket); });

    string msg;
    if (!setNonBlocking(osSocket, &msg))
    {
        m_pImpl->m_state = SocketState::Error;
        m_pImpl->m_msg = msg;
        return;
    }

    if (::connect(osSocket, res->ai_addr, res->ai_addrlen) == OSSocketError)
    {
        int err = getNetworkError();
        if (
#if PLATFORM_WINDOWS
            err != WSAEWOULDBLOCK
#else
            err != EINPROGRESS
#endif
            )
        {
            m_pImpl->m_state = SocketState::Error;
            m_pImpl->m_msg = "connect error: ";
            m_pImpl->m_msg += getNetworkErrorMsg(err);
            return;
        }

        m_pImpl->m_state = SocketState::Connecting;
    }
    else
    {
        m_pImpl->m_state = SocketState::Connected;
        m_pImpl->m_lastSendTimeMS = m_pImpl->m_lastRecvTimeMS = getTimeMS();
    }

    socketCleanup.reset();
    m_pImpl->m_osSocket = osSocket;
}

void TCPSocket::update()
{
    // Connect
    if (m_pImpl->m_state == SocketState::Connecting)
    {
        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 0;

        fd_set writeSet;
        FD_ZERO(&writeSet);
        FD_SET(m_pImpl->m_osSocket, &writeSet);
#if PLATFORM_WINDOWS
        fd_set exceptSet;
        FD_ZERO(&exceptSet);
        FD_SET(m_pImpl->m_osSocket, &exceptSet);
#endif
        int selectResult =
            select(m_pImpl->m_osSocket + 1,
                   nullptr,
                   &writeSet,
#if PLATFORM_WINDOWS
                   &exceptSet,
#else
                   nullptr,
#endif
                   &timeout);
        if (selectResult == OSSocketError)
        {
            m_pImpl->m_state = SocketState::Error;
            int err = getNetworkError();
            m_pImpl->m_msg = "select error: ";
            m_pImpl->m_msg += getNetworkErrorMsg(err);
            return;
        }

        if (selectResult > 0)
        {
            int connectErr = 0;
            socklen_t len = sizeof(connectErr);
            if (getsockopt(m_pImpl->m_osSocket,
                           SOL_SOCKET,
                           SO_ERROR,
#if PLATFORM_WINDOWS
                           (char*)&connectErr,
#else
                           (void*)&connectErr,
#endif
                           &len) == OSSocketError)
            {
                m_pImpl->m_state = SocketState::Error;
                int err = getNetworkError();
                m_pImpl->m_msg = "getsockopt error: ";
                m_pImpl->m_msg += getNetworkErrorMsg(err);
                return;
            }

            if (connectErr == 0)
            {
                m_pImpl->m_state = SocketState::Connected;
                m_pImpl->m_lastSendTimeMS =
                    m_pImpl->m_lastRecvTimeMS = getTimeMS();
            }
            else
            {
                m_pImpl->m_state = SocketState::Error;
                m_pImpl->m_msg = "Connection failed: ";
                m_pImpl->m_msg += getNetworkErrorMsg(connectErr);
                return;
            }
        }
    }

    // Receive
    if (m_pImpl->m_state == SocketState::Connected ||
        m_pImpl->m_state == SocketState::WriteClosed)
    {
        while (m_recvData.size() < 1048576)
        {
            uint8_t buf[65536];
            int bytesReceived =
                (int)recv(m_pImpl->m_osSocket,
#if PLATFORM_WINDOWS
                          (char*)buf,
                          (int)sizeof(buf),
                          0
#else
                          (void*)buf,
                          (size_t)sizeof(buf),
                          MSG_DONTWAIT
#endif
                    );
            if (bytesReceived > 0)
            {
                size_t pos = m_recvData.size();
                m_recvData.resize(pos + bytesReceived);
                memcpy(&m_recvData[pos], buf, bytesReceived);
                m_pImpl->m_lastRecvTimeMS = getTimeMS();
            }
            else if (bytesReceived == 0)
            {
                if (m_pImpl->m_state == SocketState::Connected)
                {
                    m_pImpl->m_state = SocketState::ReadClosed;
                }
                else // if (m_pImpl->m_state == SocketState::WriteClosed)
                {
                    m_pImpl->m_state = SocketState::Closed;
                }
                shutdownRead(m_pImpl->m_osSocket);
                break;
            }
            else if (bytesReceived < 0)
            {
                int err = getNetworkError();
                if (
#if PLATFORM_WINDOWS
                    err != WSAEWOULDBLOCK &&
                    err != WSAEINPROGRESS
#else
                    err != EAGAIN &&
                    err != EWOULDBLOCK &&
                    err != EINTR
#endif
                    ) 
                {
                    m_pImpl->m_state = SocketState::Error;
                    m_pImpl->m_msg = "recv error: ";
                    m_pImpl->m_msg += getNetworkErrorMsg(err);
                    return;
                }
                break;
            }
        }
    }

    // Send
    if (m_pImpl->m_state == SocketState::Connected ||
        m_pImpl->m_state == SocketState::ReadClosed)
    {
        if (!m_sendData.empty())
        {
            int bytesSent =
                (int)send(m_pImpl->m_osSocket,
#if PLATFORM_WINDOWS
                          (const char*)&m_sendData[0],
                          (int)m_sendData.size(),
                          0
#else
                          (const void*)&m_sendData[0],
                          (size_t)m_sendData.size(),
                          MSG_NOSIGNAL | MSG_DONTWAIT
#endif
                    );
            if (bytesSent > 0)
            {
                m_sendData.erase(m_sendData.begin(),
                                 m_sendData.begin() + bytesSent);
                m_pImpl->m_lastSendTimeMS = getTimeMS();
            }
            else if (bytesSent < 0)
            {
                int err = getNetworkError();
                if (
#if PLATFORM_WINDOWS
                    err != WSAEWOULDBLOCK &&
                    err != WSAEINPROGRESS
#else
                    err != EAGAIN &&
                    err != EWOULDBLOCK &&
                    err != EINTR
#endif
                    )
                {
                    m_pImpl->m_state = SocketState::Error;
                    m_pImpl->m_msg = "send error: ";
                    m_pImpl->m_msg += getNetworkErrorMsg(err);
                }
            }
        }
        else
        {
            m_pImpl->m_lastSendTimeMS = getTimeMS();
        }
    }

    // Shutdown
    if (m_pImpl->m_state == SocketState::Connected ||
        m_pImpl->m_state == SocketState::ReadClosed)
    {
        if (m_sendData.empty() &&
            m_pImpl->m_shutdownRequested)
        {
            if (m_pImpl->m_state == SocketState::Connected)
            {
                m_pImpl->m_state = SocketState::WriteClosed;
            }
            else // if (m_pImpl->m_state == SocketState::ReadClosed)
            {
                m_pImpl->m_state = SocketState::Closed;
            }
            shutdownWrite(m_pImpl->m_osSocket);
        }
    }
}

bool TCPSocket::getRemoteAddress(Address* pAddress, string* pMsg)
{
    struct sockaddr_storage addr;
    socklen_t addrlen = sizeof(addr);
    if (getpeername(m_pImpl->m_osSocket,
                    (struct sockaddr*)&addr,
                    &addrlen) == -1)
    {
        if (pMsg)
        {
            int err = getNetworkError();
            *pMsg = "getpeername: ";
            *pMsg += getNetworkErrorMsg(err);
        }
        return false;
    }

    if ((size_t)addrlen > sizeof(addr))
    {
        if (pMsg)
        {
            *pMsg = "getpeername: Buffer too small";
        }
        return false;
    }

    char hostBuf[NI_MAXHOST];
    char portBuf[NI_MAXSERV];
    int err =
        getnameinfo((const struct sockaddr*)&addr, addrlen,
                    hostBuf, sizeof(hostBuf),
                    portBuf, sizeof(portBuf),
                    NI_NUMERICHOST | NI_NUMERICSERV);
    if (err != 0)
    {
        if (pMsg)
        {
            *pMsg = "getnameinfo error: ";
#if PLATFORM_WINDOWS
            *pMsg += getNetworkErrorMsg(getNetworkError());
#else
            *pMsg += gai_strerror(err);
#endif
            return false;
        }
    }

    *pAddress = Address(hostBuf, portBuf);
    return true;
}

bool TCPSocket::getLocalAddress(Address* pAddress, string* pMsg)
{
    struct sockaddr_storage addr;
    socklen_t addrlen = sizeof(addr);
    if (getsockname(m_pImpl->m_osSocket,
                    (struct sockaddr*)&addr,
                    &addrlen) == -1)
    {
        if (pMsg)
        {
            int err = getNetworkError();
            *pMsg = "getsockname: ";
            *pMsg += getNetworkErrorMsg(err);
        }
        return false;
    }

    if ((size_t)addrlen > sizeof(addr))
    {
        if (pMsg)
        {
            *pMsg = "getsockname: Buffer too small";
        }
        return false;
    }

    char hostBuf[NI_MAXHOST];
    char portBuf[NI_MAXSERV];
    int err =
        getnameinfo((const struct sockaddr*)&addr, addrlen,
                    hostBuf, sizeof(hostBuf),
                    portBuf, sizeof(portBuf),
                    NI_NUMERICHOST | NI_NUMERICSERV);
    if (err != 0)
    {
        if (pMsg)
        {
            *pMsg = "getnameinfo error: ";
#if PLATFORM_WINDOWS
            *pMsg += getNetworkErrorMsg(getNetworkError());
#else
            *pMsg += gai_strerror(err);
#endif
            return false;
        }
    }

    *pAddress = Address(hostBuf, portBuf);
    return true;
}

void TCPSocket::close()
{
    if (m_pImpl->m_state != SocketState::Disconnected)
    {
        m_pImpl->m_shutdownRequested = true;
    }
}

void TCPSocket::reset()
{
    SocketState state = m_pImpl->m_state;

    if (state == SocketState::Disconnected)
    {
        return;
    }

    if (state == SocketState::Closed)
    {
        // Do nothing
    }
    else if (state == SocketState::WriteClosed)
    {
        shutdownRead(m_pImpl->m_osSocket);
    }
    else if (state == SocketState::ReadClosed)
    {
        shutdownWrite(m_pImpl->m_osSocket);
    }
    else
    {
        shutdownBoth(m_pImpl->m_osSocket);
    }

    closeSocket(m_pImpl->m_osSocket);

    m_sendData.clear();
    m_recvData.clear();
    m_pImpl->m_osSocket = OSInvalidSocket;
    m_pImpl->m_state = SocketState::Disconnected;
    m_pImpl->m_msg = "";
    m_pImpl->m_lastSendTimeMS = 0;
    m_pImpl->m_lastRecvTimeMS = 0;
    m_pImpl->m_shutdownRequested = false;
}

uint64_t TCPSocket::getSendStalledDurationMS() const
{
    if (m_pImpl->m_state != SocketState::Connected)
    {
        return 0;
    }

    return getTimeMS() - m_pImpl->m_lastSendTimeMS;
}

uint64_t TCPSocket::getRecvStalledDurationMS() const
{
    if (m_pImpl->m_state != SocketState::Connected)
    {
        return 0;
    }

    return getTimeMS() - m_pImpl->m_lastRecvTimeMS;
}

class TCPServerSocketImpl
{
public:
    std::vector<OSSocket> m_osSockets;
    size_t m_nextAccept;
};

TCPServerSocket::TCPServerSocket() :
    m_pImpl(new TCPServerSocketImpl)
{
    m_pImpl->m_nextAccept = 0;
}

TCPServerSocket::TCPServerSocket(TCPServerSocket&& other) :
    m_pImpl(std::move(other.m_pImpl))
{
    other.m_pImpl.reset(new TCPServerSocketImpl);
}

TCPServerSocket::~TCPServerSocket()
{
    close();
}

TCPServerSocket& TCPServerSocket::operator=(TCPServerSocket&& other)
{
    if (this != &other)
    {
        close();
        m_pImpl = std::move(other.m_pImpl);
        other.m_pImpl.reset(new TCPServerSocketImpl);
    }
    return *this;
}

bool TCPServerSocket::listen(const string& port,
                             string* pMsg)
{
    close();

    struct addrinfo hints;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags =
        AI_PASSIVE |
#if PLATFORM_WINDOWS
        8
#else
        AI_NUMERICSERV
#endif
        ;

    struct addrinfo* res = nullptr;

    {
        int err =
            getaddrinfo(nullptr,
                        port.c_str(),
                        &hints,
                        &res);

        if (err != 0)
        {
            if (pMsg)
            {
                *pMsg = "getaddrinfo error: ";
#if PLATFORM_WINDOWS
                *pMsg += getNetworkErrorMsg(getNetworkError());
#else
                *pMsg += gai_strerror(err);
#endif
            }
            return false;
        }
    }

    if (!res)
    {
        if (pMsg)
        {
            *pMsg = "getaddrinfo returned no results";
        }
        return false;
    }

    Cleanup resCleanup([&]{ freeaddrinfo(res); });

    vector<struct addrinfo*> v;
    for (uint32_t pass = 0; pass < 2; pass++)
    {
        int family =
            pass == 0 ?
            AF_INET6 :
#if PLATFORM_WINDOWS
            2
#else
            AF_INET
#endif
            ;
        for (struct addrinfo* p = res; p != nullptr; p = p->ai_next)
        {
            if (pass == 0)
            {
                char hostBuf[NI_MAXHOST];
                char portBuf[NI_MAXSERV];
                if (getnameinfo(p->ai_addr, p->ai_addrlen,
                                hostBuf, sizeof(hostBuf),
                                portBuf, sizeof(portBuf),
                                NI_NUMERICHOST | NI_NUMERICSERV) == 0)
                {
                    printf("listen: %s:%s\n", hostBuf, portBuf);
                }
            }
            if (p->ai_family == family)
            {
                v.push_back(p);
            }
        }
    }

    if (v.empty())
    {
        if (pMsg)
        {
            *pMsg = "getaddrinfo returned no suitable results";
        }
        return false;
    }

    string firstMsg;
    vector<OSSocket> osSockets;
    for (size_t i = 0; i < v.size(); i++)
    {
        struct addrinfo* p = v[i];

        OSSocket osSocket =
            socket(p->ai_family, p->ai_socktype, p->ai_protocol);
        if (osSocket == OSInvalidSocket)
        {
            if (firstMsg.empty())
            {
                int err = getNetworkError();
                firstMsg = "socket error: ";
                firstMsg += getNetworkErrorMsg(err);
            }

            continue;
        }

        Cleanup socketCleanup([&]{ closeSocket(osSocket); });

        string msg;
        if (!setNonBlocking(osSocket, &msg))
        {
            if (firstMsg.empty())
            {
                firstMsg = msg;
            }

            continue;
        }

#if !PLATFORM_WINDOWS
        if (p->ai_family == AF_INET6)
        {
            int v6only = 1;
            if (setsockopt(osSocket, IPPROTO_IPV6, IPV6_V6ONLY,
                           &v6only, sizeof(v6only)) == OSSocketError)
            {
                if (firstMsg.empty())
                {
                    int err = getNetworkError();
                    firstMsg = "setsockopt error: ";
                    firstMsg += getNetworkErrorMsg(err);
                }

                continue;
            }
        }
#endif

#if !PLATFORM_WINDOWS
        {
            int reuse = 1;
            if (setsockopt(osSocket, SOL_SOCKET, SO_REUSEADDR,
                           &reuse, sizeof(reuse)) == OSSocketError)
            {
                if (firstMsg.empty())
                {
                    int err = getNetworkError();
                    firstMsg = "setsockopt error: ";
                    firstMsg += getNetworkErrorMsg(err);
                }

                continue;
            }
        }
#endif

        if (bind(osSocket, p->ai_addr, p->ai_addrlen) == OSSocketError)
        {
            if (firstMsg.empty())
            {
                int err = getNetworkError();
                firstMsg = "bind error: ";
                firstMsg += getNetworkErrorMsg(err);
            }

            continue;
        }

        if (::listen(osSocket, SOMAXCONN) == OSSocketError)
        {
            if (firstMsg.empty())
            {
                int err = getNetworkError();
                firstMsg = "listen error: ";
                firstMsg += getNetworkErrorMsg(err);
            }

            continue;
        }

        socketCleanup.reset();
        osSockets.push_back(osSocket);
    }

    if (osSockets.empty())
    {
        if (pMsg)
        {
            *pMsg = firstMsg;
        }
        return false;
    }

    m_pImpl->m_osSockets.swap(osSockets);
    return true;
}

bool TCPServerSocket::accept(TCPSocket* pSocket, string* pMsg)
{
    if (m_pImpl->m_osSockets.empty())
    {
        if (pMsg)
        {
            *pMsg = "Not listening";
        }
        return false;
    }

    for (size_t i = 0; i < m_pImpl->m_osSockets.size(); i++)
    {
        OSSocket osListenSocket = m_pImpl->m_osSockets[m_pImpl->m_nextAccept];
        m_pImpl->m_nextAccept =
            (m_pImpl->m_nextAccept + 1) % m_pImpl->m_osSockets.size();
        OSSocket osAcceptSocket = ::accept(osListenSocket, nullptr, nullptr);
        if (osAcceptSocket == OSInvalidSocket)
        {
            int err = getNetworkError();
            if (
#if PLATFORM_WINDOWS
                err != WSAEWOULDBLOCK &&
                err != WSAEINPROGRESS
#else
                err != EAGAIN &&
                err != EWOULDBLOCK &&
                err != EINTR
#endif
                )
            {
                if (pMsg)
                {
                    *pMsg = getNetworkErrorMsg(err);
                }
                return false;
            }

            continue;
        }

        pSocket->reset();
        pSocket->m_pImpl->m_osSocket = osAcceptSocket;
        pSocket->m_pImpl->m_state = SocketState::Connected;
        pSocket->m_pImpl->m_lastSendTimeMS =
            pSocket->m_pImpl->m_lastRecvTimeMS = getTimeMS();
        return true;
    }

    pSocket->reset();
    return true;
}

void TCPServerSocket::close()
{
    for (size_t i = 0; i < m_pImpl->m_osSockets.size(); i++)
    {
        closeSocket(m_pImpl->m_osSockets[i]);
    }
    m_pImpl->m_osSockets.clear();
    m_pImpl->m_nextAccept = 0;
}
