#include "PlanSolver.hpp"
#include "Game.hpp"
#include "Map.hpp"
#include "TimeUtils.hpp"
#include <random>
#include <algorithm>

using std::vector;

#define VERBOSE 0

PlanSolver::PlanSolver(bool splurge,
                       bool test) :
    m_splurge(splurge),
    m_test(test),
    m_haveFirstMove(false),
    m_firstSource((uint32_t)-1),
    m_firstTarget((uint32_t)-1),
    m_turn((uint32_t)-1)
{
}

PlanSolver::~PlanSolver()
{
}

class PlanQueueEntry
{
public:
    PlanQueueEntry() :
        m_parentComponent(0),
        m_component(0),
        m_river(0),
        m_turns(0),
        m_options(0)
    {
    }
    PlanQueueEntry(uint32_t parentComponent,
                   uint32_t component,
                   uint32_t river,
                   uint32_t turns,
                   uint32_t options) :
        m_parentComponent(parentComponent),
        m_component(component),
        m_river(river),
        m_turns(turns),
        m_options(options)
    {
    }

public:
    uint32_t m_parentComponent;
    uint32_t m_component;
    uint32_t m_river;
    uint32_t m_turns;
    uint32_t m_options;
};

void PlanSolver::solve(const Game& game,
                       vector<uint32_t>* pRoute)
{
    if (m_haveFirstMove)
    {
        if (m_firstSource != (uint32_t)-1 &&
            m_firstTarget != (uint32_t)-1)
        {
            pRoute->push_back(m_firstSource);
            pRoute->push_back(m_firstTarget);
        }
        m_haveFirstMove = false;
        m_firstSource = (uint32_t)-1;
        m_firstTarget = (uint32_t)-1;
        return;
    }

    uint64_t stopTimeMS = getTimeMS() + 500;

    auto& map = game.m_map;
    uint32_t numSites = map.m_numSites;
    uint32_t numRivers = map.m_numRivers;
    uint32_t numMines = map.m_numMines;
    uint32_t punter = game.m_punter;
    uint32_t numPunters = game.m_numPunters;

    if (m_turn == (uint32_t)-1)
    {
        m_turn = punter;
    }
    else
    {
        m_turn += numPunters;
    }
    uint32_t turnsLeft = ((numRivers - m_turn) + numPunters - 1) / numPunters;
#if VERBOSE
    fprintf(stderr, "turn = %" PRIu32 "\n", m_turn);
    fprintf(stderr, "turnsLeft = %" PRIu32 "\n", turnsLeft);
#endif

    uint32_t optionsLeft = 0;
    if (game.m_useOptions)
    {
        optionsLeft = game.m_optionsLeft[game.m_punter];
    }

    uint32_t credits = 0;
    if (game.m_useSplurges)
    {
        uint32_t prevTurnsAllowed = m_turn / numPunters;
        uint32_t prevTurnsTaken = 0;
        for (uint32_t iRiver = 0; iRiver < numRivers; iRiver++)
        {
            if (game.m_claims[iRiver] == punter)
            {
                prevTurnsTaken++;
            }
            if (game.m_useOptions &&
                game.m_options[iRiver] == punter)
            {
                prevTurnsTaken++;
            }
        }
        if (prevTurnsAllowed > prevTurnsTaken)
        {
            credits = prevTurnsAllowed - prevTurnsTaken;
        }
    }
    uint32_t creditsLeft = credits + turnsLeft;
    if (game.m_useSplurges)
    {
#if VERBOSE
        fprintf(stderr, "credits = %" PRIu32 "\n", credits);
        fprintf(stderr, "creditsLeft = %" PRIu32 "\n", creditsLeft);
#endif
    }

    uint32_t numComponents = 0;
    vector<uint32_t> siteComponents(numSites);
    for (uint32_t iSite = 0; iSite < numSites; iSite++)
    {
        siteComponents[iSite] = (uint32_t)-1;
    }

    vector<uint32_t> queueCurrent;
    vector<uint32_t> queueNext;

    for (uint32_t iSite = 0; iSite < numSites; iSite++)
    {
        if (siteComponents[iSite] != (uint32_t)-1)
        {
            continue;
        }

        queueNext.push_back(iSite);
        while (!queueNext.empty())
        {
            queueCurrent.swap(queueNext);
            for (uint32_t iCurSite : queueCurrent)
            {
                if (siteComponents[iCurSite] != (uint32_t)-1)
                {
                    continue;
                }
                siteComponents[iCurSite] = numComponents;

                uint32_t low = map.m_siteRiverBounds[iCurSite].m_low;
                uint32_t high = map.m_siteRiverBounds[iCurSite].m_high;
                for (uint32_t iSiteRiver = low; iSiteRiver < high; iSiteRiver++)
                {
                    auto& siteRiver = map.m_siteRivers[iSiteRiver];
                    if (!(game.m_claims[siteRiver.m_river] == punter ||
                          (game.m_useOptions &&
                           game.m_options[siteRiver.m_river] == punter)))
                    {
                        continue;
                    }

                    uint32_t target = siteRiver.m_target;
                    if (siteComponents[target] == (uint32_t)-1)
                    {
                        queueNext.push_back(target);
                    }
                }
            }
            queueCurrent.clear();
        }

        numComponents++;
    }

    uint32_t numMineComponents = 0;
    vector<uint32_t> componentMineComponents(numComponents);
    for (uint32_t iComponent = 0; iComponent < numComponents; iComponent++)
    {
        componentMineComponents[iComponent] = (uint32_t)-1;
    }
    for (uint32_t iMine = 0; iMine < numMines; iMine++)
    {
        uint32_t iSite = map.m_mines[iMine];
        uint32_t iComponent = siteComponents[iSite];
        if (componentMineComponents[iComponent] != (uint32_t)-1)
        {
            continue;
        }

        componentMineComponents[iComponent] = numMineComponents;
        numMineComponents++;
    }

    vector<uint32_t> scoreTable(numMineComponents * numComponents);

    vector<uint8_t> visited(numSites);
    for (uint32_t iMine = 0; iMine < numMines; iMine++)
    {
        uint32_t iSite = map.m_mines[iMine];
        uint32_t iComponent = siteComponents[iSite];
        uint32_t iMineComponent = componentMineComponents[iComponent];

        for (uint32_t iSite = 0; iSite < numSites; iSite++)
        {
            visited[iSite] = 0;
        }

        uint32_t dist = 0;
        queueNext.push_back(iSite);
        while (!queueNext.empty())
        {
            queueCurrent.swap(queueNext);
            for (uint32_t iCurSite : queueCurrent)
            {
                if (visited[iCurSite])
                {
                    continue;
                }
                visited[iCurSite] = 1;
                uint32_t iCurComponent = siteComponents[iCurSite];
                if (iCurComponent != iComponent)
                {
                    scoreTable[iMineComponent * numComponents + iCurComponent]
                        += dist * dist;
                }

                uint32_t low = map.m_siteRiverBounds[iCurSite].m_low;
                uint32_t high = map.m_siteRiverBounds[iCurSite].m_high;
                for (uint32_t iSiteRiver = low; iSiteRiver < high; iSiteRiver++)
                {
                    auto& siteRiver = map.m_siteRivers[iSiteRiver];
                    uint32_t target = siteRiver.m_target;
                    if (!visited[target])
                    {
                        queueNext.push_back(target);
                    }
                }
            }
            queueCurrent.clear();
            dist++;
        }

        if (getTimeMS() > stopTimeMS)
        {
#if VERBOSE
            fprintf(stderr, "PlanSolver: time limit exceeded, using incomplete greedy move\n");
#endif
            break;
        }
    }

    uint32_t bestGreedyMove = (uint32_t)-1;
    uint32_t bestGreedyScore = 0;
    uint32_t bestGreedyCount = 0;

    std::default_random_engine gen{std::random_device()()};

    for (uint32_t iRiver = 0; iRiver < numRivers; iRiver++)
    {
        if (game.m_claims[iRiver] != (uint32_t)-1)
        {
            continue;
        }

        auto& river = map.m_rivers[iRiver];
        uint32_t iSourceComponent = siteComponents[river.m_source];
        uint32_t iTargetComponent = siteComponents[river.m_target];
        uint32_t iSourceMineComponent = componentMineComponents[iSourceComponent];
        uint32_t iTargetMineComponent = componentMineComponents[iTargetComponent];

        uint32_t curGreedyScore = 0;
        if (iSourceMineComponent != (uint32_t)-1)
        {
            curGreedyScore += scoreTable[iSourceMineComponent * numComponents + iTargetComponent];
        }
        if (iTargetMineComponent != (uint32_t)-1)
        {
            curGreedyScore += scoreTable[iTargetMineComponent * numComponents + iSourceComponent];
        }

        if (bestGreedyMove == (uint32_t)-1 ||
            curGreedyScore > bestGreedyScore)
        {
            bestGreedyMove = iRiver;
            bestGreedyScore = curGreedyScore;
            bestGreedyCount = 1;
        }
        else if (curGreedyScore == bestGreedyScore)
        {
            bestGreedyCount++;
            std::uniform_int_distribution<uint32_t> dist(0, bestGreedyCount - 1);
            if (dist(gen) == 0)
            {
                bestGreedyMove = iRiver;
            }
        }
    }

    if (bestGreedyMove == (uint32_t)-1)
    {
        // No legal moves -- should not happen
        return; // pass
    }

    if (bestGreedyScore == 0)
    {
        // No moves can score points for us
        // TODO: Try to prevent other teams from scoring points
        auto& river = map.m_rivers[bestGreedyMove];
        pRoute->push_back(river.m_source);
        pRoute->push_back(river.m_target);
        return;
    }

    vector<std::pair<uint32_t, uint32_t>> componentSites;
    for (uint32_t iSite = 0; iSite < numSites; iSite++)
    {
        uint32_t iComponent = siteComponents[iSite];
        componentSites.emplace_back(iComponent, iSite);
    }
    std::sort(componentSites.begin(), componentSites.end());

    vector<uint32_t> planParents(numComponents);
    vector<uint32_t> planParentMines(numComponents);
    vector<uint64_t> planScores(numComponents);
    vector<uint32_t> planRivers(numComponents);
    vector<PlanQueueEntry> planQueueCurrent;
    vector<PlanQueueEntry> planQueueNext;
    vector<PlanQueueEntry> planQueueOptions;

    uint32_t bestPlanCount = 0;
    uint64_t bestPlanRate = 0;
    uint32_t bestPlanTurns = 0;
    vector<uint32_t> bestPlanRivers;

    for (uint32_t iComponent = 0; iComponent < numComponents; iComponent++)
    {
        uint32_t iMineComponent = componentMineComponents[iComponent];
        if (iMineComponent == (uint32_t)-1)
        {
            continue;
        }

        if (getTimeMS() > stopTimeMS)
        {
#if VERBOSE
            fprintf(stderr, "PlanSolver: time limit exceeded, returning greedy move\n");
#endif
            auto& river = map.m_rivers[bestGreedyMove];
            pRoute->push_back(river.m_source);
            pRoute->push_back(river.m_target);
            return;
        }

        uint32_t bestPlanComponent = (uint32_t)-1;

        for (uint32_t iComponent = 0; iComponent < numComponents; iComponent++)
        {
            planParents[iComponent] = (uint32_t)-1;
            planParentMines[iComponent] = (uint32_t)-1;
            planScores[iComponent] = 0;
            planRivers[iComponent] = (uint32_t)-1;
        }

        planQueueOptions.emplace_back(iComponent,
                                      iComponent,
                                      (uint32_t)-1,
                                      0,
                                      0);
        while (!planQueueOptions.empty())
        {
        planQueueNext.swap(planQueueOptions);
        while (!planQueueNext.empty())
        {
            planQueueCurrent.swap(planQueueNext);
            std::shuffle(planQueueCurrent.begin(),
                         planQueueCurrent.end(),
                         gen);
            for (auto& entry : planQueueCurrent)
            {
                uint32_t iParentComponent = entry.m_parentComponent;
                uint32_t iCurComponent = entry.m_component;
                uint32_t curPlanRiver = entry.m_river;
                uint32_t curTurns = entry.m_turns;
                uint32_t curOptions = entry.m_options;
                if (planParents[iCurComponent] != (uint32_t)-1)
                {
                    continue;
                }
                planParents[iCurComponent] = iParentComponent;
                planRivers[iCurComponent] = curPlanRiver;

                if (iParentComponent == iCurComponent)
                {
                    planParentMines[iCurComponent] = iCurComponent;
                }
                else if (componentMineComponents[iParentComponent] != (uint32_t)-1)
                {
                    planParentMines[iCurComponent] = iParentComponent;
                }
                else
                {
                    planParentMines[iCurComponent] = planParentMines[iParentComponent];
                }

                if (iParentComponent != iCurComponent)
                {
                    uint64_t curPlanScore = 0;

                    bool accurate = false;
                    if (accurate)
                    {
                        {
                            uint32_t iLoopComponent = iCurComponent;
                            while (planParentMines[iLoopComponent] != iLoopComponent)
                            {
                                iLoopComponent = planParentMines[iLoopComponent];
                                uint32_t iLoopMineComponent = componentMineComponents[iLoopComponent];
                                curPlanScore += scoreTable[iLoopMineComponent * numComponents + iCurComponent];
                            }
                        }

                        uint32_t iCurMineComponent = componentMineComponents[iCurComponent];
                        if (iCurMineComponent != (uint32_t)-1)
                        {
                            uint32_t iLoopComponent = iCurComponent;
                            while (planParents[iLoopComponent] != iLoopComponent)
                            {
                                iLoopComponent = planParents[iLoopComponent];
                                curPlanScore += scoreTable[iCurMineComponent * numComponents + iLoopComponent];
                            }
                        }
                    }
                    else
                    {
                        curPlanScore += scoreTable[iMineComponent * numComponents + iCurComponent];

                        uint32_t iCurMineComponent = componentMineComponents[iCurComponent];
                        if (iCurMineComponent != (uint32_t)-1)
                        {
                            curPlanScore += scoreTable[iCurMineComponent * numComponents + iComponent];
                        }
                    }

#if 0
                    uint64_t decay = 1000;
                    for (uint32_t iTurn = 0; iTurn < curTurns - 1; iTurn++)
                    {
                        decay = (decay * 990) / 1000;
                    }
                    curPlanScore = (curPlanScore * decay) / 1000;
#endif

                    curPlanScore += planScores[iParentComponent];

                    planScores[iCurComponent] = curPlanScore;

                    uint64_t curPlanRate = 0;
                    uint32_t equivTurns = curTurns + curOptions * 1;
                    if (equivTurns == 0)
                    {
                        fprintf(stderr, "About to divide by zero -- how?\n");
                    }
                    else
                    {
                        curPlanRate = (curPlanScore * 1000) / equivTurns;
                    }

                    if (bestPlanTurns == 0 ||
                        curPlanRate > bestPlanRate)
                    {
                        bestPlanCount = 1;
                        bestPlanRate = curPlanRate;
                        bestPlanTurns = curTurns;
                        bestPlanComponent = iCurComponent;
                    }
                    else if (curPlanRate == bestPlanRate)
                    {
                        bestPlanCount++;
                        std::uniform_int_distribution<uint32_t> dist(0, bestPlanCount - 1);
                        if (dist(gen) == 0)
                        {
                            bestPlanComponent = iCurComponent;
                            bestPlanTurns = curTurns;
                        }
                    }
                }

                if (curTurns >= creditsLeft)
                {
                    continue;
                }

                auto lb = std::lower_bound(componentSites.begin(),
                                           componentSites.end(),
                                           std::pair<uint32_t, uint32_t>
                                           (iCurComponent, 0));
                auto ub = std::lower_bound(componentSites.begin(),
                                           componentSites.end(),
                                           std::pair<uint32_t, uint32_t>
                                           (iCurComponent + 1, 0));
                for (auto it = lb; it != ub; ++it)
                {
                    uint32_t iCurSite = it->second;
                    if (iCurSite >= numSites ||
                        siteComponents[iCurSite] != iCurComponent)
                    {
                        fprintf(stderr, "componentSites lookup error\n");
                        continue;
                    }

                    uint32_t low = map.m_siteRiverBounds[iCurSite].m_low;
                    uint32_t high = map.m_siteRiverBounds[iCurSite].m_high;
                    for (uint32_t iSiteRiver = low; iSiteRiver < high; iSiteRiver++)
                    {
                        auto& siteRiver = map.m_siteRivers[iSiteRiver];

                        uint32_t target = siteRiver.m_target;
                        uint32_t iTargetComponent = siteComponents[target];
                        if (planParents[iTargetComponent] != (uint32_t)-1)
                        {
                            continue;
                        }

                        uint32_t riverIdx = siteRiver.m_river;
                        if (game.m_claims[riverIdx] == (uint32_t)-1)
                        {
                            planQueueNext.emplace_back(iCurComponent,
                                                       iTargetComponent,
                                                       siteRiver.m_river,
                                                       curTurns + 1,
                                                       curOptions);
                        }
                        else if (curOptions < optionsLeft &&
                                 game.m_claims[riverIdx] != punter &&
                                 game.m_options[riverIdx] == (uint32_t)-1)
                        {
                            planQueueOptions.emplace_back(iCurComponent,
                                                          iTargetComponent,
                                                          siteRiver.m_river,
                                                          curTurns + 1,
                                                          curOptions + 1);
                        }
                    }
                }
            }
            planQueueCurrent.clear();
        }
        }

        if (bestPlanComponent != (uint32_t)-1)
        {
            bestPlanRivers.clear();
            uint32_t iCurComponent = bestPlanComponent;
            while (planParents[iCurComponent] != iCurComponent)
            {
                bestPlanRivers.push_back(planRivers[iCurComponent]);
                iCurComponent = planParents[iCurComponent];
            }
            std::reverse(bestPlanRivers.begin(), bestPlanRivers.end());
        }
    }

#if VERBOSE
    fprintf(stderr, "Best rate: %" PRIu64 " (%" PRIu32 " turns)\n",
            bestPlanRate,
            bestPlanTurns);
#endif

#if VERBOSE
    if (bestPlanTurns != (uint32_t)bestPlanRivers.size())
    {
        fprintf(stderr, "bestPlanTurns != bestPlanRivers.size()\n");
    }
#endif

    if (bestPlanRivers.empty())
    {
        // Should not happen
        return;
    }

    if (game.m_useSplurges &&
        m_splurge &&
        bestPlanTurns == (uint32_t)bestPlanRivers.size() &&
        bestPlanTurns >= 2)
    {
        vector<uint32_t> route;
        auto& firstRiver = map.m_rivers[bestPlanRivers[0]];
        route.push_back(firstRiver.m_source);
        route.push_back(firstRiver.m_target);
        auto& secondRiver = map.m_rivers[bestPlanRivers[1]];
        if (secondRiver.m_source == route[0] ||
            secondRiver.m_target == route[0])
        {
            using std::swap;
            swap(route[0], route[1]);
        }
        for (uint32_t iRiver = 1; iRiver < bestPlanTurns; iRiver++)
        {
            auto& river = map.m_rivers[bestPlanRivers[iRiver]];
            if (river.m_source == route.back())
            {
                route.push_back(river.m_target);
            }
            else if (river.m_target == route.back())
            {
                route.push_back(river.m_source);
            }
            else
            {
                break;
            }
        }
        uint32_t neededCredits = (uint32_t)route.size() - 2;
        if (credits >= neededCredits)
        {
            for (uint32_t site : route)
            {
                pRoute->push_back(site);
            }
            return;
        }
        else
        {
            return; // pass
        }
    }

    uint32_t firstRiverIdx = bestPlanRivers[0];

    if (firstRiverIdx == (uint32_t)-1)
    {
        // Should not happen
        return;
    }

    {
        auto& river = map.m_rivers[firstRiverIdx];
        pRoute->push_back(river.m_source);
        pRoute->push_back(river.m_target);
        return;
    }
}

void PlanSolver::getFutures(const Game& game,
                            vector<SolverFuture>* pFutures)
{
    if (game.m_punter != 0)
    {
        return;
    }

    vector<uint32_t> route;
    solve(game, &route);

    m_haveFirstMove = true;

    if (route.size() != 2)
    {
        m_firstSource = (uint32_t)-1;
        m_firstTarget = (uint32_t)-1;
        return;
    }

    m_firstSource = route[0];
    m_firstTarget = route[1];

    if (game.m_map.m_isMine[m_firstSource] &&
        !game.m_map.m_isMine[m_firstTarget])
    {
        pFutures->emplace_back(m_firstSource, m_firstTarget);
    }
    else if (game.m_map.m_isMine[m_firstTarget] &&
             !game.m_map.m_isMine[m_firstSource])
    {
        pFutures->emplace_back(m_firstTarget, m_firstSource);
    }
}

void PlanSolver::serialize(vector<uint32_t>* pData) const
{
    pData->push_back(m_haveFirstMove ? 1 : 0);
    pData->push_back(m_firstSource);
    pData->push_back(m_firstTarget);
    pData->push_back(m_turn);
}

bool PlanSolver::deserialize(const vector<uint32_t>& data,
                             size_t& pos)
{
    size_t size = data.size();
    if (pos + 4 > size)
    {
        return false;
    }

    m_haveFirstMove = (data[pos++] != 0);
    m_firstSource = data[pos++];
    m_firstTarget = data[pos++];
    m_turn = data[pos++];
    return true;
}
