#include "Program.hpp"
#include "Solver.hpp"
#include "NetUtils.hpp"
#include "Protocol.hpp"
#include "Convert.hpp"
#include "StringUtils.hpp"
#include "TimeUtils.hpp"
#include "MapGraphics.hpp"
#include "CalcScores.hpp"
#include <errno.h>
#include <algorithm>

using std::string;
using std::vector;

Program::Program(Solver& solver,
                 const string& punterName,
                 const string& usageName,
                 const string& extraUsage) :
    m_punterName(punterName),
    m_usageName(usageName),
    m_extraUsage(extraUsage),
    m_solver(solver),
    m_game(),
    m_online(false)
{
}

Program::~Program()
{
}

int Program::run(const vector<string>& args)
{
    if (args.size() != 0 && args.size() != 1)
    {
        usage();
        return 1;
    }

    string host;
    string port;

    if (args.size() == 1)
    {
        m_online = true;
        string hostPort(args[0]);
        size_t colonPos = hostPort.find(':');
        if (colonPos == string::npos)
        {
            host = "punter.inf.ed.ac.uk";
            port = hostPort;
        }
        else
        {
            host = hostPort.substr(0, colonPos);
            port = hostPort.substr(colonPos + 1);
        }

        if (host.empty() || port.empty())
        {
            usage();
            return 1;
        }

        for (char portChar : port)
        {
            if (portChar < '0' || portChar > '9')
            {
                usage();
                return 1;
            }
        }
    }

#if 0
    if (m_online)
    {
        fprintf(stderr, "Online mode: %s:%s\n",
                host.c_str(),
                port.c_str());
    }
    else
    {
        fprintf(stderr, "Offline mode\n");
    }
#endif

    if (m_online)
    {
        return runOnline(host, port);
    }
    else
    {
        return runOffline();
    }
}

void Program::usage()
{
    fprintf(stderr, "Usage: %s [[<host>:]<port>]\n", m_usageName.c_str());
    if (!m_extraUsage.empty())
    {
        fprintf(stderr, "%s", m_extraUsage.c_str());
    }
}

int Program::runOnline(const string& host,
                       const string& port)
{
    string msg;

    vector<Address> addresses;
    if (!lookupAddresses(host, port, &addresses, &msg))
    {
        fprintf(stderr, "%s\n", msg.c_str());
        return 1;
    }

    for (auto& address : addresses)
    {
        fprintf(stderr, "%s:%s\n",
                address.m_host.c_str(),
                address.m_port.c_str());
    }

    std::unique_ptr<MapGraphics> pMapGraphics;

    Address address = addresses[0];

    bool protocolError = false;
    bool gotColon = false;
    size_t curLen = 0;
    string curMessage;
    vector<string> messages;
    vector<string> sendMessages;

    bool gotHandshake = false;
    bool gotSetup = false;

    TCPSocket socket;
    socket.connect(address);
    bool connected = false;
    while (true)
    {
        SocketState state = socket.getState();

        if (!socket.m_recvData.empty())
        {
            for (uint8_t b : socket.m_recvData)
            {
                char ch = (char)b;
                if (!gotColon)
                {
                    if (ch == ':')
                    {
                        if (curLen == 0)
                        {
                            fprintf(stderr, "Zero or no message length\n");
                            protocolError = true;
                            break;
                        }
                        gotColon = true;
                    }
                    else if (ch < '0' || ch > '9')
                    {
                        fprintf(stderr, "Invalid message length character\n");
                        protocolError = true;
                        break;
                    }
                    else
                    {
                        curLen *= 10;
                        curLen += (ch - '0');
                    }
                }
                else
                {
                    curMessage.push_back(ch);
                    if (curMessage.size() == curLen)
                    {
                        messages.push_back(std::move(curMessage));
                        curMessage.clear();
                        curLen = 0;
                        gotColon = false;
                    }
                }
            }
            if (protocolError)
            {
                break;
            }
            socket.m_recvData.clear();

            for (auto& strMessage : messages)
            {
                if (m_online)
                {
                    fprintf(stderr, "%s", strMessage.c_str());
                    if (!strMessage.empty() &&
                        strMessage.back() != '\n')
                    {
                        fprintf(stderr, "\n");
                    }
                }

                if (!gotHandshake)
                {
                    Protocol::ServerHandshake handshake;
                    if (!parseHandshake(strMessage,
                                        &handshake,
                                        &msg))
                    {
                        fprintf(stderr, "%s\n", msg.c_str());
                        protocolError = true;
                        break;
                    }
                    gotHandshake = true;
                    fprintf(stderr, "Handshake complete\n");
                    continue;
                }

                if (!gotSetup)
                {
                    Protocol::MessageType messageType;
                    Protocol::Setup setup;
                    if (!parseMessage(true,
                                      strMessage,
                                      &messageType,
                                      &setup,
                                      nullptr,
                                      nullptr,
                                      nullptr,
                                      &msg))
                    {
                        fprintf(stderr, "%s\n", msg.c_str());
                        protocolError = true;
                        break;
                    }
                    if (messageType != Protocol::MessageType::Setup)
                    {
                        fprintf(stderr, "Expecting setup message\n");
                        protocolError = true;
                        break;
                    }
                    gotSetup = true;
                    fprintf(stderr, "Setup complete\n");
                    fprintf(stderr, "Punter %" PRIu32 "\n", setup.m_punter);
                    fprintf(stderr, "%" PRIu32 " punters\n", setup.m_punters);
                    fprintf(stderr, "%" PRIuZ " sites\n", setup.m_map.m_sites.size());
                    fprintf(stderr, "%" PRIuZ " rivers\n", setup.m_map.m_rivers.size());
                    fprintf(stderr, "%" PRIuZ " mines\n", setup.m_map.m_mines.size());

                    Protocol::Ready ready;
                    if (!processSetup(setup, &ready))
                    {
                        protocolError = true;
                        break;
                    }

                    string strReady;
                    if (!formatReady(true,
                                     setup.m_useFutures,
                                     ready,
                                     &strReady,
                                     &msg))
                    {
                        fprintf(stderr, "%s\n", msg.c_str());
                        protocolError = true;
                        break;
                    }

                    sendMessages.push_back(std::move(strReady));

                    pMapGraphics.reset(new MapGraphics(m_punterName,
                                                       m_game.m_map,
                                                       &m_game.m_claims,
                                                       &m_game.m_options,
                                                       m_game.m_punter));
                    if (!pMapGraphics->create(&msg))
                    {
                        fprintf(stderr, "%s\n", msg.c_str());
                        pMapGraphics.reset();
                    }

                    continue;
                }

                {
                    Protocol::MessageType messageType;
                    Protocol::Play play;
                    Protocol::Stop stop;
                    Protocol::Timeout timeout;
                    if (!parseMessage(true,
                                      strMessage,
                                      &messageType,
                                      nullptr,
                                      &play,
                                      &stop,
                                      &timeout,
                                      &msg))
                    {
                        fprintf(stderr, "%s\n", msg.c_str());
                        protocolError = true;
                        break;
                    }

                    if (messageType != Protocol::MessageType::Play &&
                        messageType != Protocol::MessageType::Stop &&
                        messageType != Protocol::MessageType::Timeout)
                    {
                        fprintf(stderr, "Unexpected message type (%" PRIu32 ")\n",
                                (uint32_t)messageType);
                        protocolError = true;
                        break;
                    }

                    if (messageType == Protocol::MessageType::Play)
                    {
                        fprintf(stderr, "Got play message\n");

                        Protocol::Action action;
                        if (!processPlay(play, &action))
                        {
                            protocolError = true;
                            break;
                        }

                        string strAction;
                        if (!formatAction(true, action, &strAction, &msg))
                        {
                            fprintf(stderr, "%s\n", msg.c_str());
                            protocolError = true;
                            break;
                        }

                        sendMessages.push_back(std::move(strAction));

                        if (pMapGraphics)
                        {
                            pMapGraphics->onClaims(&m_game.m_claims,
                                                   &m_game.m_options,
                                                   m_game.m_punter);
                        }
                    }

                    if (messageType == Protocol::MessageType::Stop)
                    {
                        fprintf(stderr, "Got stop message\n");

                        if (!processStop(stop))
                        {
                            protocolError = true;
                            break;
                        }

                        if (pMapGraphics)
                        {
                            pMapGraphics->onClaims(&m_game.m_claims,
                                                   &m_game.m_options,
                                                   m_game.m_punter);
                        }
                    }

                    if (messageType == Protocol::MessageType::Timeout)
                    {
                        fprintf(stderr, "Got timeout message\n");
                        fprintf(stderr, "Timeout = %f\n", timeout.m_timeout);
                    }
                }
            }
            if (protocolError)
            {
                break;
            }
            messages.clear();
        }

        if (state == SocketState::Error)
        {
            fprintf(stderr, "%s\n", socket.getMsg().c_str());
            break;
        }

        if (state == SocketState::Connected)
        {
            if (!connected)
            {
                connected = true;
                fprintf(stderr, "Connected\n");

                Address remoteAddress;
                if (socket.getRemoteAddress(&remoteAddress, &msg))
                {
                    fprintf(stderr, "Remote address: %s:%s\n",
                            remoteAddress.m_host.c_str(),
                            remoteAddress.m_port.c_str());
                }
                else
                {
                    fprintf(stderr, "Remote address: %s\n",
                            msg.c_str());
                }

                Address localAddress;
                if (socket.getLocalAddress(&localAddress, &msg))
                {
                    fprintf(stderr, "Local address: %s:%s\n",
                            localAddress.m_host.c_str(),
                            localAddress.m_port.c_str());
                }
                else
                {
                    fprintf(stderr, "Local address: %s\n",
                            msg.c_str());
                }

                Protocol::ClientHandshake handshake;
                handshake.m_me = m_punterName;

                string strHandshake;
                if (!formatHandshake(handshake,
                                     &strHandshake,
                                     &msg))
                {
                    fprintf(stderr, "%s\n", msg.c_str());
                    break;
                }

                sendMessages.push_back(std::move(strHandshake));
            }
        }

        for (auto& strMessage : sendMessages)
        {
            string strSizeMessage =
                strprintf("%" PRIuZ ":%s",
                          strMessage.size(),
                          strMessage.c_str());
            size_t pos = socket.m_sendData.size();
            size_t messageSize = strSizeMessage.size();
            socket.m_sendData.resize(pos + messageSize);
            memcpy(&socket.m_sendData[pos],
                   strSizeMessage.c_str(),
                   messageSize);
            fprintf(stderr, "%s\n", strSizeMessage.c_str());
        }
        sendMessages.clear();

        if (state == SocketState::ReadClosed)
        {
            socket.close();
        }

        if (state == SocketState::Closed)
        {
            break;
        }

        sleepMS(1);

        socket.update();

        if (pMapGraphics)
        {
            while (Graphics::update(false))
            {
                if (pMapGraphics->getDestroyed())
                {
                    break;
                }
            }

            if (pMapGraphics->getDestroyed())
            {
                break;
            }
        }
    }

    socket.reset();

    if (pMapGraphics)
    {
        //pMapGraphics->destroy();

        while (true)
        {
            sleepMS(1);

            while (Graphics::update(false))
            {
                if (pMapGraphics->getDestroyed())
                {
                    break;
                }
            }

            if (pMapGraphics->getDestroyed())
            {
                break;
            }
        }
    }

    return 0;
}

int Program::runOffline()
{
    string msg;

    {
        Protocol::ClientHandshake handshake;
        handshake.m_me = m_punterName;

        string strHandshake;
        if (!formatHandshake(handshake,
                             &strHandshake,
                             &msg))
        {
            fprintf(stderr, "%s\n", msg.c_str());
            return 1;
        }

        fprintf(stdout, "%" PRIuZ ":%s\n",
                strHandshake.size() + 1,
                strHandshake.c_str());
        fflush(stdout);
    }

    bool gotHandshake = false;

    while (true)
    {
        size_t curLen = 0;

        while (true)
        {
            int iCh = fgetc(stdin);
            if (iCh == EOF)
            {
                int e = errno;
                if (e == EAGAIN || e == EWOULDBLOCK) // for lamduct
                {
                    sleepMS(1);
                    continue;
                }
                return 1;
            }
            char ch = (char)iCh;

            if (ch == ':')
            {
                if (curLen == 0)
                {
                    fprintf(stderr, "Zero or no message length\n");
                    return 1;
                }
                break;
            }
            else if (ch < '0' || ch > '9')
            {
                fprintf(stderr, "Invalid message length character\n");
                return 1;
            }
            else
            {
                curLen *= 10;
                curLen += (ch - '0');
            }
        }

        vector<char> chars(curLen);
        size_t pos = 0;
        while (pos < curLen)
        {
            size_t bytesRead = fread(&chars[pos], 1, curLen - pos, stdin);
            if (bytesRead <= 0)
            {
                int e = errno;
                if (e == EAGAIN || e == EWOULDBLOCK) // for lamduct
                {
                    sleepMS(1);
                    continue;
                }
                fprintf(stderr, "Error reading message from stdin\n");
                return 1;
            }
            pos += bytesRead;
        }

        string strMessage(chars.begin(), chars.end());

        if (!gotHandshake)
        {
            Protocol::ServerHandshake handshake;
            if (!parseHandshake(strMessage,
                                &handshake,
                                &msg))
            {
                fprintf(stderr, "%s\n", msg.c_str());
                return 1;
            }
            gotHandshake = true;
            //fprintf(stderr, "Handshake complete\n");
            continue;
        }

        {
            Protocol::MessageType messageType;
            Protocol::Setup setup;
            Protocol::Play play;
            Protocol::Stop stop;
            Protocol::Timeout timeout;
            if (!parseMessage(false,
                              strMessage,
                              &messageType,
                              &setup,
                              &play,
                              &stop,
                              &timeout,
                              &msg))
            {
                fprintf(stderr, "%s\n", msg.c_str());
                return 1;
            }

            if (messageType != Protocol::MessageType::Setup &&
                messageType != Protocol::MessageType::Play &&
                messageType != Protocol::MessageType::Stop &&
                messageType != Protocol::MessageType::Timeout)
            {
                fprintf(stderr, "Unexpected message type (%" PRIu32 ")\n",
                        (uint32_t)messageType);
                return 1;
            }

            string sendMessage;

            if (messageType == Protocol::MessageType::Setup)
            {
                //fprintf(stderr, "Got setup message\n");

                Protocol::Ready ready;
                if (!processSetup(setup, &ready))
                {
                    return 1;
                }

                string strReady;
                if (!formatReady(false,
                                 setup.m_useFutures,
                                 ready,
                                 &strReady,
                                 &msg))
                {
                    fprintf(stderr, "%s\n", msg.c_str());
                    return 1;
                }

                sendMessage = std::move(strReady);
            }

            if (messageType == Protocol::MessageType::Play)
            {
                //fprintf(stderr, "Got play message\n");

                Protocol::Action action;
                if (!processPlay(play, &action))
                {
                    return 1;
                }

                string strAction;
                if (!formatAction(false, action, &strAction, &msg))
                {
                    fprintf(stderr, "%s\n", msg.c_str());
                    return 1;
                }

                sendMessage = std::move(strAction);
            }

            if (messageType == Protocol::MessageType::Stop)
            {
                //fprintf(stderr, "Got stop message\n");

                if (!processStop(stop))
                {
                    return 1;
                }
            }

            if (messageType == Protocol::MessageType::Timeout)
            {
                fprintf(stderr, "Got timeout message\n");
                fprintf(stderr, "Timeout = %f\n", timeout.m_timeout);
                return 1;
            }

            if (!sendMessage.empty())
            {
                fprintf(stdout, "%" PRIuZ ":%s\n",
                        sendMessage.size() + 1,
                        sendMessage.c_str());
                fflush(stdout);
            }
        }

        break;
    }

    sleepMS(50); // for lamduct

    return 0;
}

bool Program::processSetup(const Protocol::Setup& setup,
                           Protocol::Ready* pReady)
{
    m_game.m_useFutures = setup.m_useFutures;
    m_game.m_useSplurges = setup.m_useSplurges;
    m_game.m_useOptions = setup.m_useOptions;

    m_game.m_punter = setup.m_punter;
    m_game.m_numPunters = setup.m_punters;

    auto& map = m_game.m_map;

    string msg;
    if (!convertMap(setup.m_map, &map, &msg))
    {
        fprintf(stderr, "%s\n", msg.c_str());
        return false;
    }

    m_game.m_claims.resize(map.m_numRivers);
    for (uint32_t iRiver = 0; iRiver < map.m_numRivers; iRiver++)
    {
        m_game.m_claims[iRiver] = (uint32_t)-1;
    }
    if (m_game.m_useOptions)
    {
        m_game.m_options.resize(map.m_numRivers);
        for (uint32_t iRiver = 0; iRiver < map.m_numRivers; iRiver++)
        {
            m_game.m_options[iRiver] = (uint32_t)-1;
        }
    }
    m_game.initOptionsLeft();

    vector<SolverFuture> futures;
    if (m_game.m_useFutures)
    {
        m_solver.getFutures(m_game, &futures);
    }

    auto& ready = *pReady;
    ready.m_ready = m_game.m_punter;

    ready.m_futures.resize(futures.size());
    for (size_t iFuture = 0; iFuture < futures.size(); iFuture++)
    {
        ready.m_futures[iFuture].m_source = m_game.m_map.getSiteId(futures[iFuture].m_source);
        ready.m_futures[iFuture].m_target = m_game.m_map.getSiteId(futures[iFuture].m_target);
    }

    if (!m_online)
    {
        serialize(&ready.m_state);
    }

    return true;
}

bool Program::processPlay(const Protocol::Play& play,
                          Protocol::Action* pAction)
{
    if (!m_online)
    {
        if (!deserialize(play.m_state))
        {
            return false;
        }
    }

    if (!applyMoves(play.m_moves))
    {
        return false;
    }

    vector<uint32_t> route;
    uint64_t startTimeMS = getTimeMS();
    m_solver.solve(m_game, &route);
    if (m_online)
    {
        uint64_t stopTimeMS = getTimeMS();
        uint64_t solveDurationMS = stopTimeMS - startTimeMS;
        fprintf(stderr, "Solve time = %" PRIu64" ms\n", solveDurationMS);
    }

    string msg;
    bool routeOk = true;
    if (!m_game.checkAndApply(route, true, &msg))
    {
        if (m_online)
        {
            fprintf(stderr, "Solver returned illegal route: %s\n", msg.c_str());
        }
        routeOk = false;
    }

    auto& action = *pAction;
    auto& move = action.m_move;

    move.m_punter = m_game.m_punter;

    uint32_t routeSize = (uint32_t)route.size();
    if (!routeOk ||
        routeSize == 0)
    {
        move.m_moveType = Protocol::MoveType::Pass;
    }
    else
    {
        bool preferClaims = true;

        if (routeSize == 2 &&
            (preferClaims || !m_game.m_useSplurges))
        {
            move.m_moveType = Protocol::MoveType::Claim;

            if (m_game.m_useOptions)
            {
                River river(route[0], route[1]);
                uint32_t riverIdx = 0;
                if (!m_game.m_map.getRiverIdx(river, &riverIdx))
                {
                    if (m_online)
                    {
                        fprintf(stderr, "River index not found\n");
                    }
                }
                else
                {
                    if (m_game.m_claims[riverIdx] != (uint32_t)-1 &&
                        m_game.m_claims[riverIdx] != m_game.m_punter)
                    {
                        move.m_moveType = Protocol::MoveType::Option;
                    }
                }
            }

            move.m_source = m_game.m_map.getSiteId(route[0]);
            move.m_target = m_game.m_map.getSiteId(route[1]);
        }
        else
        {
            move.m_moveType = Protocol::MoveType::Splurge;
            move.m_route.resize(routeSize);
            for (uint32_t iSite = 0; iSite < routeSize; iSite++)
            {
                move.m_route[iSite] = m_game.m_map.getSiteId(route[iSite]);
            }
        }
    }

    if (!m_online)
    {
        serialize(&action.m_state);
    }

    return true;
}

bool Program::processStop(const Protocol::Stop& stop)
{
    if (!m_online)
    {
        if (!deserialize(stop.m_state))
        {
            return false;
        }
    }

    if (!applyMoves(stop.m_moves))
    {
        return false;
    }

    if (m_online)
    {
        fprintf(stderr, "Server scores:\n");
        for (auto& score : stop.m_scores)
        {
            fprintf(stderr, "%3" PRIu32 ": %" PRId64 "\n",
                    score.m_punter,
                    score.m_score);
        }

        vector<int64_t> scores;
        calcScores(m_game, &scores);

        fprintf(stderr, "Client scores:\n");
        for (uint32_t iPunter = 0; iPunter < m_game.m_numPunters; iPunter++)
        {
            fprintf(stderr, "%3" PRIu32 ": %" PRId64 "\n",
                    iPunter,
                    scores[iPunter]);
        }
    }

    return true;
}

bool Program::applyMoves(const vector<Protocol::Move>& moves)
{
    for (uint32_t iPunter = 0; iPunter < m_game.m_numPunters; iPunter++)
    {
        uint32_t desiredPunter = (m_game.m_punter + iPunter) % m_game.m_numPunters;

    for (auto& move : moves)
    {
        if (move.m_punter != desiredPunter) continue;

        if (move.m_moveType == Protocol::MoveType::Claim)
        {
            uint32_t sourceId = move.m_source;
            uint32_t targetId = move.m_target;
            uint32_t sourceIdx = 0;
            uint32_t targetIdx = 0;
            if (!m_game.m_map.getSiteIdx(sourceId, &sourceIdx) ||
                !m_game.m_map.getSiteIdx(targetId, &targetIdx))
            {
                fprintf(stderr, "Site idx lookup failed\n");
                return false;
            }
            River river(sourceIdx, targetIdx);
            uint32_t riverIdx = 0;
            if (!m_game.m_map.getRiverIdx(river, &riverIdx))
            {
                fprintf(stderr, "River idx lookup failed\n");
                return false;
            }
            if (m_game.m_claims[riverIdx] == (uint32_t)-1)
            {
                m_game.m_claims[riverIdx] = move.m_punter;
            }
            else if (m_game.m_claims[riverIdx] == move.m_punter)
            {
            }
            else
            {
                fprintf(stderr, "Conflicting claim\n");
                return false;
            }
        }

        if (move.m_moveType == Protocol::MoveType::Splurge)
        {
            uint32_t routeSize = (uint32_t)move.m_route.size();
            vector<uint32_t> route(routeSize);
            for (uint32_t iSite = 0; iSite < routeSize; iSite++)
            {
                uint32_t siteId = move.m_route[iSite];
                uint32_t siteIdx = 0;
                if (!m_game.m_map.getSiteIdx(siteId, &siteIdx))
                {
                    fprintf(stderr, "Site idx lookup failed\n");
                    return false;
                }
                route[iSite] = siteIdx;
            }
            for (uint32_t iSite = 1; iSite < routeSize; iSite++)
            {
                River river(route[iSite - 1], route[iSite]);
                uint32_t riverIdx = 0;
                if (!m_game.m_map.getRiverIdx(river, &riverIdx))
                {
                    fprintf(stderr, "River idx lookup failed\n");
                    return false;
                }
                if (m_game.m_claims[riverIdx] == (uint32_t)-1)
                {
                    m_game.m_claims[riverIdx] = move.m_punter;
                }
                else if (m_game.m_claims[riverIdx] == move.m_punter)
                {
                }
                else
                {
                    if (!m_game.m_useOptions)
                    {
                        fprintf(stderr, "Conflicting claim\n");
                        return false;
                    }

                    if (m_game.m_options[riverIdx] == (uint32_t)-1)
                    {
                        if (m_game.m_optionsLeft[move.m_punter] == 0)
                        {
                            fprintf(stderr, "No options left\n");
                            return false;
                        }
                        m_game.m_options[riverIdx] = move.m_punter;
                        m_game.m_optionsLeft[move.m_punter]--;
                    }
                    else if (m_game.m_options[riverIdx] == move.m_punter)
                    {
                    }
                    else
                    {
                        fprintf(stderr, "Conflicting option\n");
                        return false;
                    }
                }
            }
        }

        if (move.m_moveType == Protocol::MoveType::Option)
        {
            if (!m_game.m_useOptions)
            {
                fprintf(stderr, "Options not enabled\n");
                return false;
            }

            uint32_t sourceId = move.m_source;
            uint32_t targetId = move.m_target;
            uint32_t sourceIdx = 0;
            uint32_t targetIdx = 0;
            if (!m_game.m_map.getSiteIdx(sourceId, &sourceIdx) ||
                !m_game.m_map.getSiteIdx(targetId, &targetIdx))
            {
                fprintf(stderr, "Site idx lookup failed\n");
                return false;
            }
            River river(sourceIdx, targetIdx);
            uint32_t riverIdx = 0;
            if (!m_game.m_map.getRiverIdx(river, &riverIdx))
            {
                fprintf(stderr, "River idx lookup failed\n");
                return false;
            }
            if (m_game.m_claims[riverIdx] == (uint32_t)-1)
            {
                fprintf(stderr, "Option for unclaimed river\n");
                return false;
            }
            if (m_game.m_claims[riverIdx] == move.m_punter)
            {
                fprintf(stderr, "Option for own river\n");
                return false;
            }
            if (m_game.m_options[riverIdx] == (uint32_t)-1)
            {
                if (m_game.m_optionsLeft[move.m_punter] == 0)
                {
                    fprintf(stderr, "No options left\n");
                    return false;
                }
                m_game.m_options[riverIdx] = move.m_punter;
                m_game.m_optionsLeft[move.m_punter]--;
            }
            else if (m_game.m_options[riverIdx] == move.m_punter)
            {
            }
            else
            {
                fprintf(stderr, "Conflicting option\n");
                return false;
            }
        }
    }
    }

#if 1
    if (m_online)
    {
        fprintf(stderr, "Claims:");
        for (uint32_t punter : m_game.m_claims)
        {
            if (punter == (uint32_t)-1)
            {
                fprintf(stderr, " _");
            }
            else
            {
                fprintf(stderr, " %" PRIu32 "", punter);
            }
        }
        fprintf(stderr, "\n");
    }
#endif

    return true;
}

void Program::serialize(string* pState) const
{
    vector<uint32_t> data;
    m_game.serialize(&data);
    m_solver.serialize(&data);
    size_t size = data.size();

    auto& state = *pState;
    state.resize(size * 8);
    for (size_t i = 0; i < size; i++)
    {
        uint32_t val = data[i];
        for (uint32_t iDigit = 0; iDigit < 8; iDigit++)
        {
            uint32_t digit = (val >> (4 * (7 - iDigit))) & 0xf;
            char ch = digit < 10 ? ('0' + digit) : ('a' + (digit - 10));
            state[i * 8 + iDigit] = ch;
        }
    }
}

bool Program::deserialize(const string& state)
{
    if (state.size() % 8 != 0)
    {
        fprintf(stderr, "State size not a multiple of 8\n");
        return false;
    }

    size_t size = state.size() / 8;
    vector<uint32_t> data(size);
    for (size_t i = 0; i < size; i++)
    {
        uint32_t val = 0;
        for (uint32_t iDigit = 0; iDigit < 8; iDigit++)
        {
            char ch = state[i * 8 + iDigit];
            uint32_t digit = 0;
            if (ch >= '0' && ch <= '9')
            {
                digit = ch - '0';
            }
            else if (ch >= 'a' && ch <= 'f')
            {
                digit = ch - 'a' + 0xa;
            }
            else if (ch >= 'A' && ch <= 'F')
            {
                digit = ch - 'A' + 0xa;
            }
            else
            {
                fprintf(stderr, "Invalid character in state\n");
                return false;
            }
            val |= digit << (4 * (7 - iDigit));
        }
        data[i] = val;
    }

    size_t pos = 0;
    if (!m_game.deserialize(data, pos))
    {
        fprintf(stderr, "Game::deserialize() failed\n");
        return false;
    }
    if (!m_solver.deserialize(data, pos))
    {
        fprintf(stderr, "Solver::deserialize() failed\n");
        return false;
    }

    return true;
}
