#ifndef TASKQUEUE_HPP
#define TASKQUEUE_HPP

#include "Common.hpp"
#include "ThreadUtils.hpp"
#include <deque>

class TaskQueueFinished : std::exception
{
public:
    virtual const char* what() const noexcept { return "Task queue finished"; }
};

class TaskQueueEmpty : std::exception
{
public:
    virtual const char* what() const noexcept { return "Task queue empty"; }
};

template<class _TaskType>
class TaskQueue
{
public:
    typedef _TaskType TaskType;
    typedef decltype((*(TaskType*)nullptr)()) ResultType;
    typedef std::pair<TaskType, Promise<ResultType>> ValueType;

    TaskQueue() :
        m_mutex(),
        m_cond(),
        m_queue(),
        m_finished(false)
    {
    }

    ~TaskQueue() = default;

    Future<ResultType> push(TaskType task)
    {
        LockGuard<Mutex> lock(m_mutex);
        m_queue.push_back(ValueType(std::move(task), Promise<ResultType>()));
        m_cond.notifyOne();
        return m_queue.back().second.getFuture();
    }

    void pushFront(ValueType value)
    {
        LockGuard<Mutex> lock(m_mutex);
        m_queue.push_front(std::move(value));
        m_cond.notifyOne();
    }

    ValueType pop()
    {
        auto pred = [&]() { return m_finished || !m_queue.empty(); };
        UniqueLock<Mutex> lock(m_mutex);
        m_cond.wait(lock, pred);
        if (m_finished)
        {
            throw TaskQueueFinished();
        }
        ValueType value(std::move(m_queue.front()));
        m_queue.pop_front();
        return value;
    }

    ValueType tryPop()
    {
        LockGuard<Mutex> lock(m_mutex);
        if (m_finished)
        {
            throw TaskQueueFinished();
        }
        if (m_queue.empty())
        {
            throw TaskQueueEmpty();
        }
        ValueType value(std::move(m_queue.front()));
        m_queue.pop_front();
        return value;
    }

    void setFinished()
    {
        LockGuard<Mutex> lock(m_mutex);
        m_finished = true;
        m_cond.notifyAll();
    }

    bool finished() const
    {
        LockGuard<Mutex> lock(m_mutex);
        return m_finished;
    }

private:
    TaskQueue(const TaskQueue& other) = delete;
    TaskQueue(TaskQueue&& other) = delete;
    TaskQueue& operator=(const TaskQueue& other) = delete;
    TaskQueue& operator=(TaskQueue&& other) = delete;

    mutable Mutex m_mutex;
    ConditionVariable m_cond;
    std::deque<ValueType> m_queue;
    bool m_finished;
};

#endif
