#include "Common.hpp"
#include "SolverFactory.hpp"
#include "Solver.hpp"
#include "FileUtils.hpp"
#include "Protocol.hpp"
#include "Convert.hpp"
#include "Map.hpp"
#include "Game.hpp"
#include "CalcScores.hpp"
#include "TaskWorker.hpp"
#include "TaskQueue.hpp"
#include "ThreadUtils.hpp"
#include <algorithm>

using std::string;
using std::vector;
using std::pair;

class RankResult
{
public:
    vector<uint32_t> m_solverScores;
};

class RankTask
{
public:
    RankTask(const vector<pair<string, uint32_t>>& mapsIn,
             const vector<string>& solverNamesIn,
             uint32_t iMapIn,
             uint32_t iSettingsIn,
             uint32_t iOrderIn) :
        maps(mapsIn),
        solverNames(solverNamesIn),
        iMap(iMapIn),
        iSettings(iSettingsIn),
        iOrder(iOrderIn)
    {
    }

    RankResult operator()() const;

    const vector<pair<string, uint32_t>>& maps;
    const vector<string>& solverNames;
    uint32_t iMap;
    uint32_t iSettings;
    uint32_t iOrder;
};

RankResult RankTask::operator()() const
{
    uint32_t numSolvers = (uint32_t)solverNames.size();

    string strMapFile = "../maps/" + maps[iMap].first;
    uint32_t numPunters = maps[iMap].second;

    vector<std::unique_ptr<Solver>> solvers(numPunters);
    for (uint32_t iPunter = 0; iPunter < numPunters; iPunter++)
    {
        uint32_t iSolver = (iPunter + iOrder) % numSolvers;
        std::unique_ptr<Solver> pSolver =
            SolverFactory::create(solverNames[iSolver]);
        if (!pSolver)
        {
            printf("Error creating solver '%s'\n",
                   solverNames[iSolver].c_str());
            exit(1);
        }
        solvers[iPunter] = std::move(pSolver);
    }

    FILE* mapFile = fopen(strMapFile.c_str(), "rb");
    if (!mapFile)
    {
        printf("Error opening map file\n");
        exit(1);
    }

    std::vector<uint8_t> mapData;
    readFile(mapFile, &mapData);
    fclose(mapFile);
    const char* rawMapData = (const char*)&mapData[0];
    string strMapData(rawMapData, mapData.size());

    string msg;

    Protocol::Map protocolMap;
    if (!Protocol::parseMap(strMapData, &protocolMap, &msg))
    {
        printf("%s\n", msg.c_str());
        exit(1);
    }

    Game game;
    game.m_useFutures = (iSettings & 0x1) != 0;
    game.m_useSplurges = (iSettings & 0x2) != 0;
    game.m_useOptions = (iSettings & 0x4) != 0;
    
    if (!convertMap(protocolMap, &game.m_map, &msg))
    {
        printf("%s\n", msg.c_str());
        exit(1);
    }

    game.m_numPunters = numPunters;
    game.m_claims.resize(game.m_map.m_numRivers);
    for (uint32_t iRiver = 0; iRiver < game.m_map.m_numRivers; iRiver++)
    {
        game.m_claims[iRiver] = (uint32_t)-1;
    }
    if (game.m_useOptions)
    {
        game.m_options.resize(game.m_map.m_numRivers);
        for (uint32_t iRiver = 0; iRiver < game.m_map.m_numRivers; iRiver++)
        {
            game.m_options[iRiver] = (uint32_t)-1;
        }
    }
    game.initOptionsLeft();

    vector<vector<SolverFuture>> futures;
    if (game.m_useFutures)
    {
        futures.resize(numPunters);
        for (uint32_t iPunter = 0; iPunter < numPunters; iPunter++)
        {
            game.m_punter = iPunter;
            solvers[iPunter]->getFutures(game, &futures[iPunter]);
        }
    }

    for (uint32_t iMove = 0; iMove < game.m_map.m_numRivers; iMove++)
    {
        uint32_t iPunter = iMove % numPunters;
        game.m_punter = iPunter;

        vector<uint32_t> route;
        solvers[iPunter]->solve(game, &route);

        if (!game.checkAndApply(route, true, &msg))
        {
            printf("Solver returned illegal route: %s\n", msg.c_str());
        }
    }

    vector<int64_t> scores;
    calcScores(game, &scores);

    vector<pair<int64_t, uint32_t>> sortedScores(numPunters);
    for (uint32_t iPunter = 0; iPunter < numPunters; iPunter++)
    {
        sortedScores[iPunter].first = scores[iPunter];
        sortedScores[iPunter].second = iPunter;
    }
    std::sort(sortedScores.begin(), sortedScores.end());

    RankResult result;
    result.m_solverScores.resize(numSolvers);

    for (uint32_t iRank = 0; iRank < numPunters; iRank++)
    {
        uint32_t iSolver = (sortedScores[iRank].second + iOrder) % numSolvers;
        result.m_solverScores[iSolver] += iRank + 1;
    }

    return result;
}

int main(int argc, char *argv[])
{
    vector<pair<string, uint32_t>> maps
    {
        {"sample.json", 2},
        {"lambda.json", 4},
        {"circle.json", 4},
        {"Sierpinski-triangle.json", 3},
        {"randomSparse.json", 4},
        {"randomMedium.json", 4},
        {"tube.json", 8},
        {"boston-sparse.json", 8},
        {"edinburgh-sparse.json", 16},
        {"gothenburg-sparse.json", 16},
        {"nara-sparse.json", 16},
        {"oxford2-sparse-2.json", 16}
    };
    uint32_t numMaps = (uint32_t)maps.size();

    vector<string> solverNames
    {
        "plan",
        "test"
    };
    uint32_t numSolvers = (uint32_t)solverNames.size();

    vector<uint32_t> solverScores(numSolvers);

    uint32_t processorCount = Thread::hardwareConcurrency();
    printf("Number of processors: %" PRIu32 "\n",
           processorCount);

    uint32_t threadCount =
        processorCount > 0 ? processorCount : 1;
    if (threadCount < 1)
    {
        threadCount = 1;
    }
    if (threadCount > 256)
    {
        threadCount = 256;
    }
    printf("Number of threads: %" PRIu32 "\n",
           threadCount);

    TaskQueue<RankTask> taskQueue;

    vector<Thread> threads;
    for (uint32_t i = 0; i < threadCount; i++)
    {
        threads.push_back(Thread(TaskWorker<RankTask>(taskQueue)));
    }

    vector<Future<RankResult>> futures;
    
    uint32_t numRepeats = 5;
    for (uint32_t iRepeat = 0; iRepeat < numRepeats; iRepeat++)
    {
    for (uint32_t iMap = 0; iMap < numMaps; iMap++)
    {
    for (uint32_t iSettings = 0; iSettings < 8; iSettings++)
    {
    for (uint32_t iOrder = 0; iOrder < numSolvers; iOrder++)
    {
        futures.push_back(taskQueue.push(RankTask(maps,
                                                  solverNames,
                                                  iMap,
                                                  iSettings,
                                                  iOrder)));
    }
    }
    }
    }

    for (uint32_t iGame = 0; iGame < (uint32_t)futures.size(); iGame++)
    {
        RankResult result = futures[iGame].get();

        for (uint32_t iSolver = 0; iSolver < numSolvers; iSolver++)
        {
            solverScores[iSolver] += result.m_solverScores[iSolver];
        }

        printf("%6" PRIu32 ":", iGame + 1);
        for (uint32_t iSolver = 0; iSolver < numSolvers; iSolver++)
        {
            printf("    %s %9" PRIu32 "",
                   solverNames[iSolver].c_str(),
                   solverScores[iSolver]);
        }
        printf("\n");
        fflush(stdout);
    }

    taskQueue.setFinished();

    for (uint32_t i = 0; i < threadCount; i++)
    {
        threads[i].join();
    }

    return 0;
}
