#include "Common.hpp"
#include "SolverFactory.hpp"
#include "Solver.hpp"
#include "FileUtils.hpp"
#include "Protocol.hpp"
#include "Convert.hpp"
#include "Map.hpp"
#include "Game.hpp"
#include "MapGraphics.hpp"
#include "Graphics.hpp"
#include "TimeUtils.hpp"
#include "CalcScores.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: play <mapfile> <numpunters> <solver> [<solver>...]\n");
    printf("\n");
    printf("Solvers:\n");

    vector<string> nameList(SolverFactory::getList());
    for (auto& name : nameList)
    {
        printf("  %s\n", name.c_str());
    }
}

int main(int argc, char *argv[])
{
    if (argc < 4)
    {
        usage();
        return 1;
    }

    string strMapFile(argv[1]);
    string strNumPunters(argv[2]);
    vector<string> solverNames;
    uint32_t numSolvers = 0;
    for (int iArg = 3; iArg < argc; iArg++)
    {
        solverNames.push_back(argv[iArg]);
        numSolvers++;
    }

    uint32_t numPunters = (uint32_t)atoi(strNumPunters.c_str());
    if (numPunters < 1 || numPunters > 100)
    {
        usage();
        return 1;
    }

    vector<std::unique_ptr<Solver>> solvers(numPunters);
    for (uint32_t iPunter = 0; iPunter < numPunters; iPunter++)
    {
        uint32_t iSolver = iPunter % numSolvers;
        std::unique_ptr<Solver> pSolver =
            SolverFactory::create(solverNames[iSolver]);
        if (!pSolver)
        {
            usage();
            return 1;
        }
        solvers[iPunter] = std::move(pSolver);
    }

    FILE* mapFile = fopen(strMapFile.c_str(), "rb");
    if (!mapFile)
    {
        printf("Error opening map file\n");
        return 1;
    }

    std::vector<uint8_t> mapData;
    readFile(mapFile, &mapData);
    fclose(mapFile);
    const char* rawMapData = (const char*)&mapData[0];
    string strMapData(rawMapData, mapData.size());

    string msg;

    Protocol::Map protocolMap;
    if (!Protocol::parseMap(strMapData, &protocolMap, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    Game game;
    game.m_useFutures = false;
    game.m_useSplurges = false;
    game.m_useOptions = false;
    
    if (!convertMap(protocolMap, &game.m_map, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    game.m_numPunters = numPunters;
    game.m_claims.resize(game.m_map.m_numRivers);
    for (uint32_t iRiver = 0; iRiver < game.m_map.m_numRivers; iRiver++)
    {
        game.m_claims[iRiver] = (uint32_t)-1;
    }
    if (game.m_useOptions)
    {
        game.m_options.resize(game.m_map.m_numRivers);
        for (uint32_t iRiver = 0; iRiver < game.m_map.m_numRivers; iRiver++)
        {
            game.m_options[iRiver] = (uint32_t)-1;
        }
    }
    game.initOptionsLeft();

    MapGraphics mapGraphics(strMapFile,
                            game.m_map,
                            &game.m_claims,
                            &game.m_options);

    if (!mapGraphics.create(&msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    vector<uint64_t> maxSolverTimes(numPunters);

    uint64_t updateTimeMS = getTimeMS();

    vector<vector<SolverFuture>> futures;
    if (game.m_useFutures)
    {
        futures.resize(numPunters);
        for (uint32_t iPunter = 0; iPunter < numPunters; iPunter++)
        {
            game.m_punter = iPunter;
            solvers[iPunter]->getFutures(game, &futures[iPunter]);
        }
    }

    for (uint32_t iMove = 0; iMove < game.m_map.m_numRivers; iMove++)
    {
        uint32_t iPunter = iMove % numPunters;
        game.m_punter = iPunter;

        vector<uint32_t> route;
        uint64_t startTimeMS = getTimeMS();
        solvers[iPunter]->solve(game, &route);
        uint64_t stopTimeMS = getTimeMS();
        uint64_t solverTimeMS = stopTimeMS - startTimeMS;
        maxSolverTimes[iPunter] = std::max(maxSolverTimes[iPunter], solverTimeMS);
        //sleepMS(100);

        if (!game.checkAndApply(route, true, &msg))
        {
            printf("Solver returned illegal route: %s\n", msg.c_str());
        }

        mapGraphics.onClaims(&game.m_claims, &game.m_options);

        uint64_t curTimeMS = getTimeMS();
        if (curTimeMS - updateTimeMS > 20)
        {
            updateTimeMS = curTimeMS;

            while (Graphics::update(false))
            {
                if (mapGraphics.getDestroyed())
                {
                    break;
                }
            }

            if (mapGraphics.getDestroyed())
            {
                break;
            }
        }
    }

    if (mapGraphics.getDestroyed())
    {
        return 0;
    }

    vector<int64_t> scores;
    calcScores(game, &scores);

    printf("Max solver times:\n");
    for (uint32_t iPunter = 0; iPunter < numPunters; iPunter++)
    {
        printf("%3" PRIu32 ": %" PRIu64 " ms\n",
               iPunter,
               maxSolverTimes[iPunter]);
    }

    printf("Scores:\n");
    for (uint32_t iPunter = 0; iPunter < numPunters; iPunter++)
    {
        printf("%3" PRIu32 ": %" PRId64 "\n",
               iPunter,
               scores[iPunter]);
    }

    while (true)
    {
        //mapGraphics.destroy();

        sleepMS(20);

        while (Graphics::update(false))
        {
            if (mapGraphics.getDestroyed())
            {
                break;
            }
        }

        if (mapGraphics.getDestroyed())
        {
            break;
        }
    }

    return 0;
}
